#include "generic_simulator.h"


/**
 * Constructor
 */
Simulator::Simulator(std::string host):
    num_resets(0),
    last_reward(0),
    accumulated_reward(0)
{
    // load config file
    config_reader.init("rexd_config.cfg");
    utils::init_loggers("logger.conf");
    
    if (!config_reader.get_variable<bool>("simulator_use_std_cin")) {
        uint port_vr = config_reader.get_variable<uint>("simulator_port");
        if (port_vr == 0) { // auto selection
            port_vr = utils::get_socket_from_current_dir_hash();
        }
        std::string port = boost::lexical_cast<std::string>(port_vr);

        std::cout << "Connecting to host " << host << "... ";
        tcp_client.connect(host, port);
        if (!tcp_client)
        {
            std::cerr << "Unable to connect to host " << host << " on port " << port << std::endl;
            exit(-1);
        }
        client_istream = &tcp_client;
        client_ostream = &tcp_client;
        std::cout << "connected!" << std::endl;
    }
    else {
        client_ostream = &std::cout;
        client_istream = &std::cin;
    }

    srand((unsigned)time(NULL));

    // load simulation files
    scenario_ptr = boost::make_shared<Scenario>(config_reader.get_path("ppddl_domain_file_ground_truth"), 
                                                config_reader.get_path("simulator_ppddl_current_problem_file"), 
                                                config_reader);
    std::cout << "\033[1;34m" << "Reset! (" << ++num_resets << ")" << "\033[0m" << std::endl;
}

void Simulator::reset(const std::string& problem_path)
{
    last_reward = 0;
    accumulated_reward = 0;
    scenario_ptr->reset_problem_file(problem_path);
    scenario_ptr->write_PDDL_problem(config_reader.get_path("simulator_ppddl_current_problem_file"));
}

void Simulator::reset_parser(const std::string& input_command)
{
    std::string reset_num = input_command.substr(input_command.find("reset")+5,1);
    if (!utils::is_number(reset_num)) reset_num = "";
    std::cout << reset_num;
    std::stringstream ss;
    ss << config_reader.get_path("simulator_ppddl_init_problem_file") << reset_num << ".pddl";
    reset(ss.str());
}


/**
 * Main infinite loop
 */
void Simulator::main_loop()
{
    bool continue_simulation = true;
    while (*client_istream && continue_simulation) {
        std::string input_command;
        std::getline(*client_istream, input_command);
        continue_simulation = parse_command(input_command);
    }

    std::cout << "Disconnected." << std::endl;
}


bool Simulator::parse_command(const std::string& input_command)
{
    std::cout << " Input command is: " << input_command << std::endl;
    if (input_command.size() < 3) {
        std::cout << "Ignoring too short command" << std::endl;
    }
    else if (input_command.find("request_state")!=std::string::npos) {
        *client_ostream << get_state() << std::endl << get_objects() << std::endl;
        *client_ostream << last_reward << std::endl;
    }
    else if (input_command.find("planning_started")!=std::string::npos) {
//         std::cout << "Planning started" << std::endl;
    }
    else if (input_command.find("planning_finished")!=std::string::npos) {
//         std::cout << "Planning finished" << std::endl;
    }
    else if (input_command.find("reset")!=std::string::npos) { // last one because reset is included in reset2
        std::cout << "\033[1;34m" << "Reset";
        reset_parser(input_command);
        std::cout << "! (" << ++num_resets << ")" << "\033[0m" << std::endl;
    }
    else if (input_command.find("finish_simulator")!=std::string::npos) {
        std::cout << "Finishing simulation..." << std::endl;
        return false;
    }
    else {
        std::string result;
        // + triggers teacher action
        if (input_command.find("+")!=std::string::npos) {
            std::cout << "Received action " << "\033[1;35m" <<  input_command << "\033[0m" << std::endl;
            if (config_reader.get_variable<bool>("teacher_can_fail")) {
                result = apply_action(Symbol(input_command), false);
            }
            else {
                result = apply_action(Symbol(input_command), true);
            }
        }
        else {
            std::cout << "Received action " << input_command << std::endl;
            result = apply_action(Symbol(input_command), false);
        }
        *client_ostream << "action_finished_" << result << std::endl << std::endl;
    }
    return true;
}


/**
 * Get state
 */
std::string Simulator::get_state()
{
    std::stringstream ss;

    ss << scenario_ptr->get_state();

    return ss.str();
}

std::string Simulator::get_objects()
{
    std::stringstream ss;
    ss << "(:objects ";
    BOOST_FOREACH(const PDDLTypeObjectMap::value_type& type_objects, scenario_ptr->get_state().pddl_object_manager.get_ppddl_objects_per_type()) {
        BOOST_FOREACH(const std::string object, type_objects.second) {
            ss << object << " ";
        }
        ss << "- " << type_objects.first << " ";
    }
    ss << ")";

    return ss.str();
}

/**
 * Add action
 */
std::string Simulator::apply_action(Symbol action, bool forced_outcome)
{
    bool success, applied;
    float rule_reward = 0;
    applied = scenario_ptr->apply_action(action, success, rule_reward, forced_outcome);
    last_reward = rule_reward;
    accumulated_reward += last_reward;
    scenario_ptr->write_PDDL_problem(config_reader.get_path("simulator_ppddl_current_problem_file"));
    if (applied) {
        if (success) {
            std::cout << "\033[1;32m" << "Action " << action << " was applied successfully" << " ( " << last_reward << " | " << accumulated_reward << " )" << "\033[0m" << std::endl;
            return "success";
        }
        else
            std::cout << "\033[1;33m" << "Action " << action << " was applied NOT successfully" << " ( " << last_reward << " | " << accumulated_reward << " )" << "\033[0m" << std::endl;
        return "applied_not_success";
    }
    else {
        std::cout << "\033[1;31m" << "Action " << action << " couldn't be applied" << " ( " << last_reward << " | " << accumulated_reward << " )" << "\033[0m" << std::endl;
        return "failed";
    }
}



int main(int argc, char** argv) {
    if (argc >= 3)
    {
        std::cerr << "Usage: simulator [host]" << std::endl;
        return 1;
    }

    if (argc >= 2)
    {
        if ((std::string(argv[1]) == "-h") || (std::string(argv[1]) == "--help")) {
            std::cerr << "Usage: simulator [host]" << std::endl;
            return 1;
        }
    }

    std::string host;
    if (argc >= 2)
        host = argv[1];
    else
        host = "127.0.0.1";

    Simulator simulator(host);

    simulator.main_loop();
    return 0;
}

