#include <boost/timer.hpp>

#include <symbolic/rules.h>
#include <symbolic/predicates.h>
#include "symbolic/scenario.h"

#include "lfd/failure_analyzer.h"
#include "rexd/planner_factory.h"
#include "rexd/pasula_learner.h"
#include "rexd/rexd_scenario.h"



#define LOGA(Level_, Logger_, ...)   \
    while (Log::Enabled(Level_)) {  \
      Log_Write(Level_, Logger_, __func__, __FILE__, __LINE__, __VA_ARGS__);  \
      break;  \
    }


void Log_Write(uint level,
               char const* logger,
               char const* function,
               char const* file,
               int line,
               std::string msg)
{
    std::cout << msg;
}


void compare_rules(char* rules1_path, char* rules2_path) {
    RuleSet rules1;
    rules1.read_NID_rules(rules1_path);
    RuleSet rules2;
    rules2.read_NID_rules(rules2_path);

    rules1.remove_nop_rules();
    rules2.remove_nop_rules();
//     LOG(INFO) << "Rules1: " << rules1;
//     LOG(INFO) << "Rules2: " << rules2;
    std::sort (rules1.begin(), rules1.end());
    std::sort (rules2.begin(), rules2.end());

//     LOG(INFO) << "Rules1: " << rules1;
//     LOG(INFO) << "Rules2: " << rules2;

    if (rules1 < rules2) {
        std::cout << START_COLOR_RED << "Rules are different!!! rules1 > rules2" << END_COLOR << std::endl;
        std::cout << "Differences: \n" << rules1.get_different_rules(rules2) << std::endl;
    }
    else if (rules2 < rules1) {
        std::cout << START_COLOR_RED << "Rules are different!!! rules2 > rules1" << END_COLOR << std::endl;
        std::cout << "Differences: \n" << rules1.get_different_rules(rules2) << std::endl;
    }
    else {
        std::cout << "Rules are equal!" << std::endl;
    }
}


void compare_rules_score(char* rules1_path, char* rules2_path) {
    RuleSet rules1;
    rules1.read_NID_rules(rules1_path);
    RuleSet rules2;
    rules2.read_NID_rules(rules2_path);
    TransitionGroup transitions = read_transitions_from_file("transitions.dat");

    float score = rules1.score(transitions);
    LOG(INFO)  << "Rules 1 score: " << score;

    score = rules2.score(transitions);
    LOG(INFO)  << "Rules 2 score: " << score;
}


void learn(std::string action_string = "")
{
    struct timeval timer_start = utils::start_timer();
    LOG(INFO) << "Learning";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);
    
    LearnerPtr learner = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));
    DomainPtr domain_ptr = DomainPtr(new Domain);
    TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
    
    if ((config_reader.get_variable<bool>("learning")) && (transitions.size() > 0) ) {
        CINFO("planning") << START_COLOR_BLUE << "Learning" << END_COLOR;
        RuleSetPtr removed_rules_ptr, added_rules_ptr;
        RuleSetPtr rules_ptr;
        if (action_string != "") {
            rules_ptr = learner->learn_action_rules(*rules_ptr, transitions, TypedSymbol(action_string, std::vector<std::string>(), std::vector<std::string>()), removed_rules_ptr, added_rules_ptr);
        }
        else {
            rules_ptr = learner->learn(transitions);
        }
        rules_ptr->write_rules_to_file(config_reader.get_path("nid_rules_path"));
        domain_ptr->index_rules(*rules_ptr);
    }
    else {
        CINFO("planning") << "Skipping learning (no transitions or no previous action)";
    }
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
}


void migrate_domain()
{
    LOG(INFO) << "Migrating domain...";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    // reward
    GoalGroup goals("reward.dat");
    
    // current scenario
    State current_state;
    current_state.read_from_file("state_simulator.dat");
    
    // init scenario
    State init_state;
    init_state.read_from_file("./simulated_init_state.dat");
    
    // current rules
    RuleSetPtr current_rules_ptr(new RuleSet);
    current_rules_ptr->read_NID_rules(config_reader.get_path("nid_rules_path"));
    
    // grounded domain
    RuleSetPtr ground_truth_rules_ptr(new RuleSet);
    ground_truth_rules_ptr->read_NID_rules(config_reader.get_path("ground_truth_nid_rules_path"));

    // finish action
    if (config_reader.get_variable<bool>("final_reward_action")) {
        std::vector<TypedSymbol> action_list = ground_truth_rules_ptr->get_actions();
        std::vector<TypedSymbol>::iterator finish_action_it = std::find(action_list.begin(), action_list.end(), MAKE_TYPED_SYMBOL_FINISH_ACTION);
        if (finish_action_it == action_list.end()) { // if finish_action wasn't found, load it manually
            RuleSet finish_rule;
            finish_rule.read_NID_rules(config_reader.get_path("nid_finish_rule_path"));
            ground_truth_rules_ptr->add_rules(finish_rule);
            current_rules_ptr->add_rules(finish_rule);
            // -finished() predicate used by final reward action
            ground_truth_rules_ptr->add_predicate_to_preconditions(Predicate<Symbol>(MAKE_SYMBOL_FINISHED, false));
            current_rules_ptr->add_predicate_to_preconditions(Predicate<Symbol>(MAKE_SYMBOL_FINISHED, false));
        }
    }
    
    Scenario current_scenario(*current_rules_ptr, current_state, goals, config_reader);
//     if (!current_rules_ptr->empty()) {
        current_scenario.write_PDDL_domain(config_reader.get_path("ppddl_domain_file"));
//     }
    if (!current_state.empty()) {
        current_scenario.write_PDDL_problem(config_reader.get_path("simulator_ppddl_current_problem_file"));
    }
    
    Scenario ground_truth_scenario(*ground_truth_rules_ptr, init_state, goals, config_reader);
    if (!ground_truth_rules_ptr->empty()) {
        ground_truth_scenario.write_PDDL_domain(config_reader.get_path("ppddl_domain_file_ground_truth"));
    }
    if (!init_state.empty()) {
        ground_truth_scenario.write_PDDL_problem(config_reader.get_path("simulator_ppddl_init_problem_file") + ".pddl");
    }
}


void generate_problem_with_incomplete_state(Scenario scenario_to_modify, const uint file_number, const PlannerPtr planner)
{
    uint DEBUG = 1;
    PredicateGroup<Symbol>::const_iterator predicate_it;
    do {
        predicate_it = utils::get_random_element<PredicateGroup<Symbol>::const_iterator>(scenario_to_modify.get_state().begin(), scenario_to_modify.get_state().end());
    } while (predicate_it->is_special()); // obtain non-special symbol
    
    scenario_to_modify.get_modificable_state().update_predicate(predicate_it->get_negated());
    CLOG_IF(DEBUG > 1, INFO, "excuses") << "Changing " << *predicate_it << " to " << predicate_it->get_negated();
    
    // check if plan can be found
    float planning_result;
    RexdScenario rexd_scenario(scenario_to_modify);
    rexd_scenario.set_vmin();
    SymbolOptional new_res = planner->plan(rexd_scenario, planning_result, true, false);
    if ( (!new_res) || (*new_res == MAKE_TYPED_SYMBOL_TEACHER) ) { // if no solution, then save the scenario to find excuses
        CLOG_IF(DEBUG > 0, INFO, "excuses") << ".";
        scenario_to_modify.write_PDDL_problem("simulator_ppddl_current_problem_file_" + boost::lexical_cast<std::string>(file_number) + ".pddl");
        utils::write_string_to_file("result_problem_" + boost::lexical_cast<std::string>(file_number) + ".txt", boost::lexical_cast<std::string>(predicate_it->get_negated()));
    }
}


void generate_problem_with_wrong_rules(Scenario scenario_to_modify, const uint file_number, const PlannerPtr planner)
{
    uint DEBUG = 1;
    RulePtr random_rule;
    do {
        random_rule = *utils::get_random_element<RuleSet::iterator>(scenario_to_modify.get_modificable_rules().begin(), scenario_to_modify.get_modificable_rules().end());
    } while (*random_rule == MAKE_TYPED_SYMBOL_FINISH_ACTION); // Don't change finish action!
    
    PredicateGroup<Symbol> solution;
    if ((std::rand() % 2) == 0) { // 50% add random precondition
        CLOG_IF(DEBUG > 1, INFO, "excuses") << "Adding precondition " << NOENDL;
        PredicateGroup<Symbol> predicates = scenario_to_modify.get_state();
        predicates.make_symbolic();
        predicates.remove_duplicated();
        Predicate<Symbol> random_predicate = *utils::get_random_element<PredicateGroup<Symbol>::iterator>(predicates.begin(), predicates.end());
        if ((std::rand() % 2) == 0) { // 50% true
            random_predicate.positive = true;
        }
        else {
            random_predicate.positive = false;
        }
        random_rule->preconditions.update_predicate(random_predicate);
        solution.push_back(random_predicate);
    }
    else { // 50% remove random outcome predicate
        CLOG_IF(DEBUG > 1, INFO, "excuses") << "Removing outcome " << NOENDL;
        OutcomeGroup::iterator random_outcome_it = utils::get_random_element<OutcomeGroup::iterator>(random_rule->outcomes.begin(), random_rule->outcomes.end());
        if (!random_outcome_it->empty()) { //if a predicate to be removed exists
            PredicateGroup<Symbol>::iterator random_outcome_predicate_it = utils::get_random_element<PredicateGroup<Symbol>::iterator>(random_outcome_it->begin(), random_outcome_it->end());
            solution.push_back(*random_outcome_predicate_it); // has to be added before it is removed!
            random_outcome_it->erase(random_outcome_predicate_it);
        }
    }
    CLOG_IF(DEBUG > 1, INFO, "excuses") << solution;
    
    // check if plan can be found
    float planning_result;
    RexdScenario rexd_scenario(scenario_to_modify);
    rexd_scenario.set_vmin();
    SymbolOptional new_res = planner->plan(rexd_scenario, planning_result, true, false);
    if ( (!new_res) || (*new_res == MAKE_TYPED_SYMBOL_TEACHER) ) { // if no solution, then save the scenario to find excuses
        CLOG_IF(DEBUG > 0, INFO, "excuses") << "*";
        scenario_to_modify.write_PDDL_domain("ppddl_domain_file_" + boost::lexical_cast<std::string>(file_number) + ".pddl");
        utils::write_string_to_file("result_domain_" + boost::lexical_cast<std::string>(file_number) + ".txt", boost::lexical_cast<std::string>(solution));
    }
}


void generate_set_of_problems_with_planning_failures()
{
    LOG(INFO) << "Generating domain problems with planning failures";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    RuleSetPtr rules_ptr;
    CINFO("planning") << "Using ground truth rules";
    PDDLFileReader pddl_file_reader(config_reader.get_path("ppddl_domain_file_ground_truth"));
    pddl_file_reader.parse_domain_file(rules_ptr);

    RexdScenario scenario(Scenario(*rules_ptr, config_reader.get_path("simulator_ppddl_current_problem_file"), config_reader));
    scenario.set_vmin();
    
    uint file_number = 100;
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);
    while (!scenario.is_in_goal()) {
        for (uint i = 0; i < 10; i++ ) {
            generate_problem_with_incomplete_state(scenario, file_number + i, planner);
            generate_problem_with_wrong_rules(scenario, file_number + i, planner);
        }
        file_number += 100;
        
        float planning_result;
        SymbolOptional new_res = planner->plan(scenario, planning_result, true, true);
        if (new_res) {
            scenario.apply_action(*new_res);
        }
        else {
            CERROR("planning") << "Couldn\'t obtain a plan to reach the goal.";
        }
    }
}


void plan(int n_actions)
{
    struct timeval timer_start = utils::start_timer();

    LOG(INFO) << "Planning";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    RuleSetPtr rules_ptr;
    PDDLFileReader pddl_file_reader(config_reader.get_path("ppddl_domain_file"));
    pddl_file_reader.parse_domain_file(rules_ptr);
    if (rules_ptr->size() < 2) {
        CWARNING("planning") << "Using ground truth rules";
        PDDLFileReader pddl_file_reader(config_reader.get_path("ppddl_domain_file_ground_truth"));
        pddl_file_reader.parse_domain_file(rules_ptr);
    }

    RexdScenario scenario(Scenario(*rules_ptr, config_reader.get_path("simulator_ppddl_current_problem_file"), config_reader));
    scenario.set_vmin();
    
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);
    if (n_actions > 1) {
        bool reached_goal;
        std::vector<Symbol> new_res = planner->plan_n_actions(scenario, n_actions, reached_goal, true, true);
    }
    else {
        float planning_result;
        SymbolOptional new_res = planner->plan(scenario, planning_result, true, true);
    }
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
}

std::string read_from_stream(std::istream& input_stream)
{
    std::string result;
    std::getline (input_stream, result);
    return result;
}

SymbolOptional read_action()
{
    utils::send_notification_msg("REX-D", "Requiring teacher demonstration! ");
    std::cout << "Write action to apply (or \"finish_action()\" to exit): " << std::flush;
    std::string action_string = read_from_stream(std::cin);
    if (action_string.empty()) {
        return SymbolOptional();
    }
    else {
        return SymbolOptional(Symbol(action_string));
    }
}

void interactive_plan()
{
    struct timeval timer_start = utils::start_timer();

    LOG(INFO) << "Planning";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    RuleSetPtr rules_ptr;
    PDDLFileReader pddl_file_reader(config_reader.get_path("ppddl_domain_file"));
    pddl_file_reader.parse_domain_file(rules_ptr);
    if (rules_ptr->size() < 2) {
        CWARNING("planning") << "Using ground truth rules";
        PDDLFileReader pddl_file_reader(config_reader.get_path("ppddl_domain_file_ground_truth"));
        pddl_file_reader.parse_domain_file(rules_ptr);
    }

    RexdScenario scenario(Scenario(*rules_ptr, config_reader.get_path("simulator_ppddl_current_problem_file"), config_reader));
    scenario.set_vmin();
    
    LOG(INFO) << "Current state is " << scenario.get_state();
    SymbolOptional new_action = read_action();
    while (new_action && (*new_action != MAKE_TYPED_SYMBOL_FINISH_ACTION)) {
        scenario.apply_action(*new_action);
        LOG(INFO) << "Current state is " << scenario.get_state();
        new_action = read_action();
    }
    
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
}

void explore()
{
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    RuleSet rules;
    rules.read_NID_rules(config_reader.get_path("nid_rules_path"));
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);

    State state;
    state.read_from_file(config_reader.get_path("simulator_current_state_path"));
    TransitionGroup transitions = read_transitions_from_file("transitions_full.dat");

    std::vector<TypedSymbol> action_list;
    if (!config_reader.get_variable<bool>("teacher")) {
        LOG(ERROR) << "TODO!!!!";
        assert(0);
        // TODO
//         read_action_list("symbols.dat");
    }
    else {
        action_list = rules.get_actions();
    }
    Scenario scenario(rules, state, GoalGroup(), config_reader, action_list);
    Explorer explorer(scenario, boost::make_shared<TransitionGroup>(transitions));
    explorer.is_state_known();
}


SymbolOptional plan_with_subgoals(const RuleSet& rules, const State& state, FailureAnalyzer& excuse_generator, const PredicateGroup<Symbol>& excuse, PlannerPtr planner, const Goal& goal, const ConfigReader& config_reader)
{
    // TODO FIX
//     Goal new_goal = excuse_generator.get_subgoal(goal, excuse);
//     GoalGroup new_goals;
//     new_goals.push_back(new_goal);
//     CINFO("planning") << "Planning using subgoals with excuse " << excuse;
//     CINFO("planning") << "New goal is " << new_goal;
//     Scenario scenario(rules, state, new_goals, config_reader);
//     SymbolOptional result = planner->plan(scenario, false);
//     return result;
    return SymbolOptional();
}


void generate_excuses()
{
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);
    RuleSet rules;
    rules.read_NID_rules(config_reader.get_path("nid_rules_path"));
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);

    State state;
    state.read_from_file(config_reader.get_path("simulator_current_state_path"));
    GoalGroup goals = GoalGroup(config_reader.get_path("goal_file"));

    std::vector< Symbol > planning_res;
    RuleSet grounded_rules = rules.get_pddl_grounded_rules(state);
    grounded_rules.remove_nop_rules();
    LOG(INFO) << "************************* GENERATING EXCUSES *************************";

    Scenario scenario(rules, state, goals, config_reader);
    FailureAnalyzer excuse_generator(scenario);

    PredicateGroup<Symbol> excuse = excuse_generator.get_planning_failure();
    LOG(INFO)  << "Best excuse is " << excuse;
    
    PredicateGroup<Symbol> symbolic_excuse(excuse);
    symbolic_excuse.make_symbolic();
    std::vector<std::string> actions_requiring_excuse = rules.find_rules_with_a_precondition(excuse).get_action_names();

    LOG(INFO) << "Actions requiring excuse: " << actions_requiring_excuse;

    SymbolOptional action_opt;
    if (config_reader.get_variable<bool>("rule_analysis_equivalent_rules")) {
        CINFO("excuses") << "Looking for equivalent rules!";
        TransitionGroup transitions = read_transitions_from_file("transitions_full.dat");
//         float expected_reward;
//         action_opt = excuse_generator.search_for_equivalent_useful_rules(excuse, transitions, boost::make_shared<RuleSet>(rules), expected_reward);
        // TODO
        LOG(ERROR) << "TODO " << "Rule alternatives";
    }

    // predicate subgoals are incompatible with vmin
    if ((config_reader.get_variable<bool>("rule_analysis_subgoals")) && (!config_reader.get_variable<bool>("use_rmax")) ) {
        CINFO("excuses") << "Trying subgoals";
        action_opt = plan_with_subgoals(rules, state, excuse_generator, excuse, planner, goals.get_predicates(), config_reader);
    }
}


void print_help()
{
    std::cout << "Usage: rexd_utils [option] [args]" << std::endl
              << "Options:" << std::endl
              << "\texcuses \t\t- Plans with excuses using current folder files" << std::endl
              << "\tplan \t\t\t- plans using current folder files" << std::endl
              << "\texplore \t\t- explores using current folder files" << std::endl
              << "\trules <rules1> <rules2>\t- compares both rule sets" << std::endl
              << "\tscore <rules1> <rules2>\t- gets score of both rule sets" << std::endl
              << "\tmigrate \t\t- migrates scenario from NID to PPDDL" << std::endl
              << "\tinteractive \t\t- interactively solve the problem (no planning)" << std::endl
              << "\tfailures \t\t- generate random failures for the domain" << std::endl;
}


int main(int argc, char** argv) {
    // generated random seed
    srand((unsigned)time(0));
    utils::init_loggers("logger.conf");

    if (argc < 2) {
        print_help();
    }
    else {
        std::string aux(argv[1]);
        if (aux == "excuses") {
            utils::init_loggers("logger.conf");
            generate_excuses();
        }
        else if (aux == "plan") {
            utils::init_loggers("logger.conf");
            if (argc >= 3)
                plan(atoi(argv[2]));
            else
                plan(1);
        }
        else if (aux == "learn") {
            utils::init_loggers("logger.conf");
            if (argc > 2) {
                learn(argv[2]);
            }
            else {
                learn();
            }
        }
        else if (aux == "explore") {
            utils::init_loggers("logger.conf");
            explore();
        }
        else if (aux == "rules") {
            compare_rules(argv[2], argv[3]);
        }
        else if (aux == "score") {
            compare_rules_score(argv[2], argv[3]);
        }
        else if (aux == "migrate") {
            migrate_domain();
        }
        else if (aux == "interactive") {
            interactive_plan();
        }
        else if (aux == "failures") {
            generate_set_of_problems_with_planning_failures();
        }
        else {
            print_help();
        }
    }

    return 0;
}
