 
#include <iostream>
#include <algorithm>
#include <boost/iostreams/device/mapped_file.hpp>

#include <gtest/gtest.h>

#include <test_helpers.h>

bool compare_two_files(const std::string& file1, const std::string& file2)
{
    boost::iostreams::mapped_file_source f1(file1);
    boost::iostreams::mapped_file_source f2(file2);

    if(    f1.size() == f2.size()
        && std::equal(f1.data(), f1.data() + f1.size(), f2.data())
       )
    {
//         std::cout << "The files are equal\n";
        return true;
    }
    else {
//         std::cout << "The files are not equal\n";
        return false;
    }
}

Scenario load_scenario(const ConfigReader& config_reader)
{
    return Scenario(config_reader.get_path("ppddl_domain_file"), config_reader.get_path("simulator_ppddl_current_problem_file"), config_reader);
}

std::vector<std::string> get_actions_requiring_excuse(const Scenario& scenario, FailureAnalyzer& failure_analyzer)
{
    std::vector<std::string> actions_requiring_excuse;
    
    PredicateGroup<Symbol> excuse = failure_analyzer.get_planning_failure();
    if (!excuse.empty()) {
        RuleSet rules_requiring_excuse = scenario.get_rules().find_rules_with_a_precondition(excuse);
        rules_requiring_excuse.remove_nop_rules();
        actions_requiring_excuse = rules_requiring_excuse.get_action_names();
    }
    
    return actions_requiring_excuse;
}

ListPredicateGroup::ListPredicateGroup(std::string line) 
{
    // remove leading and trailing spaces
    line = utils::trim(line);

    std::stringstream ss(line);

    std::string item;
    while (std::getline(ss, item, '|')) {
        item = utils::trim(item);

        if (item.size() > 1) { 
            push_back(PredicateGroup<Symbol>(item));
        }
    }
}
    
bool ListPredicateGroup::has_predicate_group(const PredicateGroup<Symbol>& predicates) const
{
    bool result = false;
    
    BOOST_FOREACH(const PredicateGroup<Symbol>& predicates_it, *this) {
        if (predicates_it == predicates) {
            result = true;
        }
    }
    
    return result;
}


void test_excuses(const std::string& folder, const ListPredicateGroup& possible_expected_excuses, const bool full_test, const std::string& action_containing_excuse)
{
        std::cout << std::endl;
        LOG(INFO) << "TEST: " << folder;
        // load config
        ConfigReader config_reader;
        config_reader.init(folder + "/rexd_config.cfg", folder);
        if (!full_test) {
            // as these test were copied automatically with a script, template path has to be set manually
            config_reader.set_variable<std::string>("ppddl_rules_template_path", "../../../../../../scenarios/generic/template_ppddl");
        }
        
        // load high_level_planner
//         TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
        Scenario scenario = load_scenario(config_reader);
        FailureAnalyzer excuse_generator(scenario);
        PredicateGroup<Symbol> excuse = excuse_generator.get_planning_failure();
        std::vector<std::string> actions_requiring_excuse = get_actions_requiring_excuse(scenario, excuse_generator);
        LOG(INFO) << "Excuse: " << excuse;
        LOG(INFO) << "Actions requiring excuse: " << actions_requiring_excuse;
        
        // Check excuse
        EXPECT_EQ(possible_expected_excuses.has_predicate_group(excuse), true)
            << "Obtained excuse " << excuse << ", when best excuse is " << possible_expected_excuses.front() << ". Test: " << folder;
        
        if (full_test) {
            // Check actions required by excuse
            if (action_containing_excuse == "") {
                EXPECT_EQ(true, actions_requiring_excuse.empty())
                    << "There should be no actions requiring excuse, instead of " << actions_requiring_excuse << ". Test: " << folder;
            }
            else {
                EXPECT_EQ(true, utils::belongs_to_vector<std::string>(action_containing_excuse, actions_requiring_excuse))
                    << "Actions requiring excuse should have " << action_containing_excuse << ", but instead they are " << actions_requiring_excuse << ". Test: " << folder;
            }
        }
}


void test_excuses(const std::string& domain, const std::string& problem)
{
    boost::filesystem::path p(boost::filesystem::current_path());
    
    p = p / "excuse_tests" / domain / problem;
    
    if (!boost::filesystem::is_directory(p)) {
        LOG(ERROR) << p << " does not exist!";
        return;
    }
    else {
        boost::filesystem::directory_iterator base_iter(p);
        boost::filesystem::directory_iterator end;
        for ( ; base_iter !=end; ++base_iter) { // "domain" and "problem" subfolders
            boost::filesystem::directory_iterator problem_iter(base_iter->path());
            for ( ; problem_iter !=end; ++problem_iter) {
                std::string problem_folder = boost::lexical_cast<std::string>(problem_iter->path());
                boost::replace_all(problem_folder, "\"", "");
                LOG(INFO) << "Folder: " << problem_iter->path();
                std::string expected_result_str = utils::get_file_contents(problem_folder + "/result.txt");
                ListPredicateGroup expected_result(expected_result_str);
                test_excuses(problem_folder, expected_result, false);
            }
        }
    }
}

void test_excuses(const FolderExcusePairs::value_type& folder_excuse)
{
    std::string test_folder_full = "excuse_tests/" + folder_excuse.first + "/";
    test_excuses(test_folder_full, folder_excuse.second.first, true, folder_excuse.second.second);
}

