/*
 * Various code tests
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// external
#include <boost/assign/std/vector.hpp>
#include <boost/assign/list_inserter.hpp>

// rexd
#include "symbolic/predicates.h"
#include "symbolic/rules.h"
#include "symbolic/state.h"
#include "rexd/rexd_scenario.h"
#include "rexd/pasula_learner.h"
#include <rexd/high_level_planner.h>

// helpers
#include "test_helpers.h"


TEST(Predicates, Satisfying) { 
    PredicateGroup<Symbol> predicates1("plate(71) plate(72) plate(73) cup(74) onTable(71) onTable(72) onTable(73) onTable(74) clear(71) clear(72) clear(73) clear(74) -finished()");
    PredicateGroup<Symbol> predicates2("plate(71) plate(72) plate(73) cup(74) onTable(71) onTable(72) onTable(73) onTable(74) clear(71) clear(72) clear(73) clear(74)");
    PredicateGroup<Symbol> predicates3("plate(71) plate(72) plate(73) cup(74) onTable(71) onTable(72) onTable(73) onTable(74) clear(71) clear(72) clear(73) clear(74) finished()");
    PredicateGroup<Symbol> predicates4("plate(71) plate(72) plate(73) cup(74) onTable(71) onTable(72) onTable(73) onTable(74) clear(71) clear(72) clear(73) clear(74) -finished()");
    PredicateGroup<Symbol> predicates5("-broken(71) -broken(72) -broken(73) -broken(74) -broken(75) -broken(76) -broken(77) -onTable(71) -onTable(72) -onTable(73) -onTable(74) -onTable(75)");
    PredicateGroup<Symbol> predicates6("-broken(71) -broken(72) -broken(73) -broken(74) -broken(75) -broken(76) -broken(77)");
    PredicateGroup<Symbol> predicates7("-plate(71) plate(72) plate(73)");
    PredicateGroup<Symbol> predicates8;
    
    // satisfies
    ASSERT_EQ( true, predicates1.satisfies(PredicateGroup<Symbol>("plate(71)").front()) );
    ASSERT_EQ( 0, predicates1.satisfies(PredicateGroup<Symbol>("-plate(71)").front()) );
    ASSERT_EQ( 0, predicates1.satisfies(PredicateGroup<Symbol>("plate(74)").front()) );
    
    // is_satisfied_by
    ASSERT_EQ( true,  predicates2.is_satisfied_by(predicates1) );
    ASSERT_EQ( 0, predicates3.is_satisfied_by(predicates1) );
    ASSERT_EQ( true,  predicates4.is_satisfied_by(predicates1) );
    ASSERT_EQ( 0, predicates5.is_satisfied_by(predicates1) );
    ASSERT_EQ( true,  predicates6.is_satisfied_by(predicates1) );
    ASSERT_EQ( 0, predicates7.is_satisfied_by(predicates1) );
    ASSERT_EQ( true,  predicates8.is_satisfied_by(predicates1) );
    
    ASSERT_EQ( 12, predicates2.count_satisfied_by(predicates1) );
    ASSERT_EQ( 12, predicates3.count_satisfied_by(predicates1) );
    ASSERT_EQ( 13, predicates4.count_satisfied_by(predicates1) );
    ASSERT_EQ( 8,  predicates5.count_satisfied_by(predicates1) );
    ASSERT_EQ( 7,  predicates6.count_satisfied_by(predicates1) );
    ASSERT_EQ( 2,  predicates7.count_satisfied_by(predicates1) );
    ASSERT_EQ( 0,  predicates8.count_satisfied_by(predicates1) );
    
    ASSERT_EQ( PredicateGroup<Symbol>(""),              predicates2.get_not_satisfied_by(predicates1) );
    ASSERT_EQ( PredicateGroup<Symbol>("finished()"),    predicates3.get_not_satisfied_by(predicates1) );
    ASSERT_EQ( PredicateGroup<Symbol>(""),              predicates4.get_not_satisfied_by(predicates1) );
    ASSERT_EQ( PredicateGroup<Symbol>("-onTable(71) -onTable(72) -onTable(73) -onTable(74)"), predicates5.get_not_satisfied_by(predicates1) );
    ASSERT_EQ( PredicateGroup<Symbol>(""),              predicates6.get_not_satisfied_by(predicates1) );
    ASSERT_EQ( PredicateGroup<Symbol>("-plate(71)"),    predicates7.get_not_satisfied_by(predicates1) );
    ASSERT_EQ( PredicateGroup<Symbol>(""),              predicates8.get_not_satisfied_by(predicates1) );
}


TEST(RulesTest, GroundingRules) { 
    PredicateGroup<Symbol> predicates("clear(62), -clear(63), -on(62,71), -onTable()");
    RuleSet symbolic_rules;
    symbolic_rules.read_NID_rules("rules_tests/rules_symbolic_basic.dat");
    RuleSet grounded_rules;
    grounded_rules.read_NID_rules("rules_tests/rules_grounded_basic.dat");
//     RuleSet grounded_symbolic_rules = symbolic_rules.get_grounded_rules(predicates);
    
    static const std::string objects_arr[] = {"62","63","71"};
    std::vector<std::string> objects_vec (objects_arr, objects_arr + sizeof(objects_arr) / sizeof(objects_arr[0]) );
    static const std::string types_arr[] = {"default","default","default"};
    std::vector<std::string> types_vec (types_arr, types_arr + sizeof(types_arr) / sizeof(types_arr[0]) );
    
    State state(predicates);
    state.pddl_object_manager.set_ppddl_objects_from_object_type_pairs(objects_vec, types_vec);
    
    EXPECT_EQ(grounded_rules, symbolic_rules.get_pddl_grounded_rules(state, false)) 
        << "Grounding symbolic rules without deictic references failed.";
}
 
// TEST(RulesTest, DeicticGroundingRules) {
//     State state("clear(62), -clear(63), on(71,63), on(72,62), on(72,63), -onTable()");
//     
//     LOG(WARNING) << "Deictic references not supported any more.";
//     
//     static const std::string objects_arr[] = {"62","63","71","72"};
//     std::vector<std::string> objects_vec (objects_arr, objects_arr + sizeof(objects_arr) / sizeof(objects_arr[0]) );
//     static const std::string types_arr[] = {"default","default","default"};
//     std::vector<std::string> types_vec (types_arr, types_arr + sizeof(types_arr) / sizeof(types_arr[0]) );
//     state.pddl_object_manager.set_ppddl_objects_from_object_type_pairs(objects_vec, types_vec);
//     
//     RuleSet symbolic_rules;
//     symbolic_rules.read_NID_rules("rules_tests/rules_symbolic_deictic.dat");
//     RuleSet grounded_rules;
//     grounded_rules.read_NID_rules("rules_tests/rules_grounded_deictic.dat");
//     
//     EXPECT_EQ(grounded_rules, symbolic_rules.get_pddl_grounded_rules(state))
//         << "Grounding symbolic rules with deictic references failed.";
// }
// 
// TEST(RulesTest, DeicticGroundingRulesCorner) {
//     utils::init_loggers("generic/logger.conf");
//     
//     LOG(WARNING) << "Deictic references not supported any more.";
//     
//     State state("-vehicleAt(71) road(71,72) road(72,73) road(71,81) road(72,82) road(81,72) road(82,73) spareIn(81) spareIn(82) road(81,91) road(91,82) spareIn(91) -notFlattire() -hasspare() vehicleAt(81)");
//     
//     static const std::string objects_arr[] = {"71","72","73","81","82","91"};
//     std::vector<std::string> objects_vec (objects_arr, objects_arr + sizeof(objects_arr) / sizeof(objects_arr[0]) );
//     static const std::string types_arr[] = {"default","default","default"};
//     std::vector<std::string> types_vec (types_arr, types_arr + sizeof(types_arr) / sizeof(types_arr[0]) );
//     state.pddl_object_manager.set_ppddl_objects_from_object_type_pairs(objects_vec, types_vec);
//     
//     RuleSet symbolic_rules;
//     symbolic_rules.read_NID_rules("rules_tests/rules_symbolic_deictic_corner_case.dat");
//     RuleSet grounded_rules;
//     grounded_rules.read_NID_rules("rules_tests/rules_grounded_corner_case.dat");
//     
//     SymbolicRule sym_rule = *symbolic_rules.front();
//     RuleSet grounded_rules2 = sym_rule.get_pddl_grounded_rules(state, true);
// //     LOG(INFO) << "Grounded rules are " << grounded_rules2;
//     
//     EXPECT_EQ(grounded_rules, symbolic_rules.get_pddl_grounded_rules(state))
//         << "Grounding symbolic rules with deictic references corner case failed.";
// }
// 
// TEST(RulesTest, DeicticGroundingRulesCoversTransitionPreconditions) { 
//     TransitionGroup transitions = read_transitions_from_file("rules_tests/transitions_rules_symbolic_deictic_covering.dat");
//     RuleSet symbolic_rules;
//     symbolic_rules.read_NID_rules("rules_tests/rules_symbolic_deictic_covering.dat");
//     State state("base(70) -at(70) loc(70) -onGround() humanAlive() loc(71) loc(72) loc(73) loc(74) -finished() -at(72) -at(71) at(74) -at(73)");
//     
//     static const std::string objects_arr[] = {"70","71","72","73","74"};
//     std::vector<std::string> objects_vec (objects_arr, objects_arr + sizeof(objects_arr) / sizeof(objects_arr[0]) );
//     static const std::string types_arr[] = {"default","default","default"};
//     std::vector<std::string> types_vec (types_arr, types_arr + sizeof(types_arr) / sizeof(types_arr[0]) );
//     state.pddl_object_manager.set_ppddl_objects_from_object_type_pairs(objects_vec, types_vec);
//     
//     ASSERT_EQ( 14, symbolic_rules.front()->count_covering_transitions_preconditions(transitions));
// }

TEST(DomainTest, GeneratingDomain) {
    ConfigReader config_reader;
    config_reader.init("domain_tests/rexd_config.cfg");
    State state;
    state.read_from_file("domain_tests/domain_state.dat");
    RuleSet rules;
    rules.read_NID_rules("domain_tests/domain_rules.dat");
    GoalGroup goals("domain_tests/domain_goals.dat");
    Scenario scenario(rules, state, goals, config_reader);
    std::stringstream ss;
    scenario.domain.write_NID(ss);
    utils::write_string_to_file("generated/symbols.dat", ss.str());
    EXPECT_EQ(true, compare_two_files("domain_tests/domain_symbols.dat", "generated/symbols.dat"))
        << "Domain does not create expected symbols file.";
}

TEST(PPDDLTest, GeneratingPPDDLRules) {
    ConfigReader config_reader;
    config_reader.init("domain_tests/rexd_config.cfg");
    State state;
    state.read_from_file("domain_tests/domain_state.dat");
    RuleSet rules;
    rules.read_NID_rules("domain_tests/domain_rules.dat");
    GoalGroup goals("domain_tests/domain_goals.dat");
    RexdScenario rexd_scenario(Scenario(rules, state, goals, config_reader));
    
    rexd_scenario.write_IPPC_PPDDL_to_file("./domain_tests/template_ppddl", "generated/ppddl_rules.ppddl");
    
    EXPECT_EQ(true, compare_two_files("domain_tests/ppddl_rules.ppddl", "generated/ppddl_rules.ppddl"))
        << "Error generating PPDDL rules file.";
}

TEST(ExploreTest, CheckKnown) 
{
    utils::init_loggers("generic/logger.conf");
    
    LOG(ERROR) << "UPDATE TEST FILES. OLD TRANSITION FILES CANNOT BE USED ANY MORE.";
    ASSERT_EQ(true,false);
    
    using namespace boost::assign;
    typedef std::map<std::string, bool> FolderActionBoolPairs;
    FolderActionBoolPairs test_folders; 
    insert( test_folders )
        ( "empty_known"     , false )
        ( "explored"        , false )
        ( "1action_unknown" , true  )
        ( "almost_explored" , true  )
        ( "almost_explored2", true  )
        ( "almost_explored3", true  );
    
    BOOST_FOREACH(FolderActionBoolPairs::value_type folder_action, test_folders) {
        // load config
        std::string test_folder_full = "exploration_tests/" + folder_action.first + "/";
        ConfigReader config_reader;
        config_reader.init(test_folder_full + "/rexd_config.cfg", test_folder_full);
        
        // load high_level_planner
        std::ofstream logging_file;
        logging_file.open ("/dev/null", std::ios::app);
        HighLevelPlanner high_level_planner(config_reader, &logging_file);
        PredicateGroup<Symbol> excuse;
        std::vector<std::string> actions_requiring_excuse;
        TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
        Scenario scenario = load_scenario(config_reader);
        
        // test exploration without forced flag
        SymbolOptional explore_action = high_level_planner.explore(scenario, transitions, false);
        EXPECT_EQ(folder_action.second, explore_action.is_initialized())
            << "State should be known (no action instead of " << *explore_action << "). Test: " << folder_action.first;
        
        // test exploration with forced flag
        if (scenario.get_rules().empty()) { 
            explore_action = high_level_planner.explore(scenario, transitions, true);
            SymbolOptional ground_truth_action(Symbol("placeshaft()"));
            EXPECT_EQ(0, explore_action.is_initialized())
                << "State should be known (no action instead of " << *explore_action << "). Test: " << folder_action.first;
        }
        else {
            explore_action = high_level_planner.explore(scenario, transitions, true);
            SymbolOptional ground_truth_action(Symbol("placeshaft()"));
            EXPECT_EQ(true, explore_action.is_initialized())
                << "A exploration action should have been selected (because forced flag). Test: " << folder_action.first;
        }
    }
}

TEST(ExploreTest, ActionSelection) 
{
    utils::init_loggers("generic/logger.conf");
    
    LOG(ERROR) << "UPDATE TEST FILES. OLD TRANSITION FILES CANNOT BE USED ANY MORE.";
    ASSERT_EQ(true,false);
    
    using namespace boost::assign;
    typedef std::map<std::string, SymbolOptional> FolderActionPairs;
    FolderActionPairs test_folders; 
    insert( test_folders )
        ( "1action_unknown" , SymbolOptional(Symbol("placeshaft()"))       )
        ( "almost_explored" , SymbolOptional(Symbol("screwpendulumhead()")))
        ( "almost_explored2", SymbolOptional(Symbol("placependulum()"))    )
        ( "almost_explored3", SymbolOptional(Symbol("placepeg4()"))        );
    
    BOOST_FOREACH(FolderActionPairs::value_type folder_action, test_folders) {
        // load config
        std::string test_folder_full = "exploration_tests/" + folder_action.first + "/";
        ConfigReader config_reader;
        config_reader.init(test_folder_full + "/rexd_config.cfg", test_folder_full);
        
        // load high_level_planner
        std::ofstream logging_file;
        logging_file.open ("/dev/null", std::ios::app);
        HighLevelPlanner high_level_planner(config_reader, &logging_file);
        PredicateGroup<Symbol> excuse;
        std::vector<std::string> actions_requiring_excuse;
        TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
        Scenario scenario = load_scenario(config_reader);
        
        // test exploration without forced flag
        SymbolOptional explore_action = high_level_planner.explore(scenario, transitions, false);
        SymbolOptional ground_truth_action(Symbol("placeshaft()"));
        EXPECT_EQ(true, explore_action.is_initialized())
                << "A exploration action should have been selected. Test: " << folder_action.first;
        EXPECT_EQ(folder_action.second, explore_action)
            << "Exploring " << *explore_action << " instead of " << *folder_action.second << ". Test: " << folder_action.first;
        
        // test exploration with forced flag
        explore_action = high_level_planner.explore(scenario, transitions, true);
        EXPECT_EQ(true, explore_action.is_initialized())
                << "A exploration action should have been selected. Test: " << folder_action.first;
        EXPECT_EQ(folder_action.second, explore_action)
            << "Exploring " << *explore_action << " instead of " << *folder_action.second << ". Test: " << folder_action.first;
        
        // test exploration in E3 exploration/exploitation algorithm
        RexdScenario rexd_scenario = load_scenario(config_reader);
        explore_action= high_level_planner.exploit_or_explore(rexd_scenario, transitions, excuse, actions_requiring_excuse);
        EXPECT_EQ(true, explore_action.is_initialized())
                << "A exploration action should have been selected. Test: " << folder_action.first;
        EXPECT_EQ(folder_action.second, explore_action)
            << "Exploring " << *explore_action << " instead of " << *folder_action.second << ". Test: " << folder_action.first;
    }
}


TEST(ExcuseTest, GenerateExcusesAutas) {
    utils::init_loggers("generic/logger.conf");
    using namespace boost::assign;
    FolderExcusePairs test_folders; 
    insert( test_folders )
        ( "autas_p06_1"   , ExcuseActionPair(ListPredicateGroup("freeFbHb1()"), "placependulum")     )
        ( "autas_p06_2"   , ExcuseActionPair(ListPredicateGroup("freeFbHb2()"), "screwpendulumhead") )
        ( "autas_horiz_2" , ExcuseActionPair(ListPredicateGroup("clearB1()")  , "placepeg1") 
        );
    
    BOOST_FOREACH(FolderExcusePairs::value_type folder_excuse, test_folders) {
        test_excuses(folder_excuse);
    }
}

TEST(ExcuseTest, GenerateExcusesExBlock) {
    utils::init_loggers("generic/logger.conf");
    using namespace boost::assign;
    FolderExcusePairs test_folders; 
    insert( test_folders )
        ( "exblock_p5_1" , ExcuseActionPair(ListPredicateGroup("on(63, 66)"),          "")           )
        ( "exblock_p5_2" , ExcuseActionPair(ListPredicateGroup("on(63,66) on(65,63)"), "")           )
        ( "exblock_p5_3" , ExcuseActionPair(ListPredicateGroup("-onTable(66)"),        "putOnBlock") )
        ( "exblock_p5_4" , ExcuseActionPair(ListPredicateGroup("-onTable(66)"),        "putOnBlock") /*)
        ( "exblock_dangerous_1" , ExcuseActionPair(ListPredicateGroup("noDestroyedTable()"), "putDown")*/
        );
    
    BOOST_FOREACH(FolderExcusePairs::value_type folder_excuse, test_folders) {
        test_excuses(folder_excuse);
    }
}

TEST(ExcuseTest, GenerateExcusesKitchen) {
    utils::init_loggers("generic/logger.conf");
    using namespace boost::assign;
    FolderExcusePairs test_folders; 
    insert( test_folders )
        ( "kitchen_p1_1", ExcuseActionPair(ListPredicateGroup("-onTable(71) -onTable(72) -onTable(73) -onTable(74)"),               "")           )
        ( "kitchen_p1_2", ExcuseActionPair(ListPredicateGroup("-broken(71) -cup(74) clear(73)"),                                    "putPlateOn") )             
        ( "kitchen_p1_3", ExcuseActionPair(ListPredicateGroup("clear(71) -cup(74) | clear(71) plate(74)"),                          "putPlateOn") )
        ( "kitchen_p1_4", ExcuseActionPair(ListPredicateGroup("clear(71) | -cup(74)"),                                              "putPlateOn") )
        ( "kitchen_p1_5", ExcuseActionPair(ListPredicateGroup("clear(71) | clear(72) | cup(71) | cup(72)"),                         "putPlateOn") ) 
        ( "kitchen_p1_6", ExcuseActionPair(ListPredicateGroup("-broken(73) cup(71)"),                                               "putPlateOn") )
        ( "kitchen_p1_7", ExcuseActionPair(ListPredicateGroup("-broken(75) cup(71) | -broken(75) cup(73) | -broken(75) clear(73)"), "putPlateOn") )
        ( "kitchen_p1_8", ExcuseActionPair(ListPredicateGroup("clear(71) | clear(72)"),                                             "putPlateOn") ) 
        ( "kitchen_p1_9", ExcuseActionPair(ListPredicateGroup("plate(74)"),                                                         "putPlateOn")
        );
    
    BOOST_FOREACH(FolderExcusePairs::value_type folder_excuse, test_folders) {
        test_excuses(folder_excuse);
    }
}

TEST(ExcuseTest, GenerateExcusesAutasDet) {
    utils::init_loggers("generic/logger.conf");
    test_excuses("autas","det");
}

TEST(ExcuseTest, GenerateExcusesKitchenP1) {
    utils::init_loggers("generic/logger.conf");
    test_excuses("kitchen","p1reset");
}

TEST(ExcuseTest, GenerateExcusesTriangle) {
    utils::init_loggers("generic/logger.conf");
    
    using namespace boost::assign;
    FolderExcusePairs test_folders; 
    insert( test_folders )
        ( "triangle_p1_1"    , ExcuseActionPair(ListPredicateGroup("notFlattire()"), "moveCar") )
        ( "triangle_p1_2"    , ExcuseActionPair(ListPredicateGroup("notFlattire()"), "moveCar") )
        ( "triangle_p1_3"    , ExcuseActionPair(ListPredicateGroup("notFlattire()"), "moveCar") )
        ( "triangle_p1_4"    , ExcuseActionPair(ListPredicateGroup("spareIn(72)()"), "loadtire") )
        
        ( "triangle_p1vmin_1", ExcuseActionPair(ListPredicateGroup("vehicleAt(75)"), "") )
        ( "triangle_p1vmin_2", ExcuseActionPair(ListPredicateGroup("spareIn(75)"),   "moveCar") )
        ( "triangle_p1vmin_3", ExcuseActionPair(ListPredicateGroup("notFlattire()"), "moveCar") )
        ( "triangle_p1vmin_4", ExcuseActionPair(ListPredicateGroup("notFlattire()"), "moveCar") )
        ( "triangle_p1vmin_5", ExcuseActionPair(ListPredicateGroup("notFlattire()"), "moveCar") )
        ( "triangle_p1vmin_6", ExcuseActionPair(ListPredicateGroup("spareIn(72)"),   "loadtire") 
        );
    
    BOOST_FOREACH(FolderExcusePairs::value_type folder_excuse, test_folders) {
        test_excuses(folder_excuse);
    }
}


TEST(LearnerTest, PasulaAllRulesDeterministic) {
    utils::init_loggers("generic/logger.conf");
    
    LOG(ERROR) << "UPDATE TEST FILES. OLD TRANSITION FILES CANNOT BE USED ANY MORE.";
    ASSERT_EQ(true,false);
    
    ConfigReader config_reader;
    config_reader.init("domain_tests/rexd_config.cfg");
    TransitionGroup transitions = read_transitions_from_file("domain_tests/transitions.dat");
    RuleSet ground_truth_rules;
    ground_truth_rules.read_NID_rules("domain_tests/learned_nid_rules_all.dat");
    RuleSetPtr ground_truth_rules_ptr = boost::make_shared<RuleSet>(ground_truth_rules);
    RuleSet ground_truth_action_rules;
    ground_truth_action_rules.read_NID_rules("domain_tests/learned_nid_rules_action.dat");
    RuleSetPtr ground_truth_action_rules_ptr = boost::make_shared<RuleSet>(ground_truth_action_rules);
    LearnerPtr learner_ptr = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));
    
    RuleSetPtr learned_rules_ptr = learner_ptr->learn_quiet(transitions);
    // set same indexes as ground truth, else files won't be equal
    // rule index cannot be restarted, static variable in constructor
    learned_rules_ptr->set_identifiers(ground_truth_rules_ptr);
    
    RuleSetPtr learned_action_rules_ptr = learner_ptr->learn_action_rules(*ground_truth_rules_ptr, transitions, TypedSymbol("placefaceplatefront()"));
    
    learned_rules_ptr->write_rules_to_file("generated/learned_nid_rules.dat");
    learned_action_rules_ptr->write_rules_to_file("generated/learned_nid_action_rules.dat");
    
    EXPECT_EQ(*ground_truth_rules_ptr, *learned_rules_ptr)
        << "Error learning all rules.";
    EXPECT_EQ(*ground_truth_action_rules_ptr, *learned_action_rules_ptr)
        << "Error learning action rules.";
    EXPECT_EQ(true, compare_two_files("domain_tests/learned_nid_rules_all.dat", "generated/learned_nid_rules.dat"))
        << "Error generating NID rules file.";
    EXPECT_EQ(true, compare_two_files("domain_tests/learned_nid_rules_action.dat", "generated/learned_nid_action_rules.dat"))
        << "Error generating NID rules file.";
}
 
int main(int argc, char **argv) {
    // argv parsed by google tests
    // Remember: --gtest_filter=*Excuse* to execute just excuse tests
    testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}