#include "ippc_wrapper/ippc_planner.h"

IppcPlanner::IppcPlanner(const ConfigReader& config_reader)
{
    ppddl_rules_file = config_reader.get_path("ppddl_grounded_rules_path");
    ppddl_template_file = config_reader.get_path("ppddl_rules_template_path");
    planning_miliseconds = config_reader.get_variable<uint>("planning_time");
    planning_fast_miliseconds = config_reader.get_variable<uint>("planning_fast_time");
    planning_epsilon = config_reader.get_variable<double>("planning_epsilon");
    planner_path = config_reader.get_binary_path("planner_path");
    planner_algorithm = config_reader.get_variable<std::string>("planner_algorithm");
    planner_port = config_reader.get_variable<uint>("planner_port");
    if (planner_port == 0) { // auto selection
        planner_port = utils::get_socket_from_current_dir_hash(1);
    }
//     ippcserver.start_session();
//     ippcserver.start_round();
}

IppcPlanner::~IppcPlanner()
{
//     ippcserver.end_round();
//     ippcserver.end_session();
}


int IppcPlanner::estimate_distance_to_finish(const RexdScenario &rexd_scenario)
{
    bool reached_goal;
    std::vector< Symbol > planning_res = plan_n_actions(rexd_scenario, rexd_scenario.config_reader.get_variable<uint>("planning_horizon"), reached_goal, false, false);
    std::cerr << "IppcPlanner::estimate_distance_to_finish: " << "Estimation: " << planning_res.size() << std::endl;
    return planning_res.size();
}


std::vector< Symbol > IppcPlanner::plan_n_actions(const RexdScenario& rexd_scenario, const uint n_actions, bool &reached_goal, const bool thoroughly, const bool verbose, const bool force_effect) const
{
//     int DEBUG = 0;
    std::vector< Symbol > result;
    RexdScenario updated_scenario(rexd_scenario);
    SymbolOptional previous_action_opt;
    SymbolOptional current_action_opt;
    float previous_reward=0.0;
    reached_goal = false;

    if (rexd_scenario.is_in_goal()) {
        reached_goal = true;
        if (verbose)
            CINFO("planning") << "Is goal. Not planning.";
        return result;
    }

    for (uint i = 0; i < n_actions; ++i) {
        if (verbose)
            CINFO("planning") << " Plan action " << i << " with state: " << updated_scenario.get_state();
        else
            std::cout << "." << std::flush;
        
        // plan
        float current_reward;
        current_action_opt = plan(updated_scenario, current_reward, thoroughly, verbose);
        updated_scenario.decrease_planning_horizon();

        // check if plan is improving
        assert((i==0) || previous_action_opt);
        if (previous_action_opt) { // always true but first time
            double planning_reward_step = updated_scenario.config_reader.get_variable<float>("planning_reward_step");
            if ((previous_reward - current_reward) < planning_reward_step) { // improving
                result.push_back( *previous_action_opt );
            }
            else { // not improving
                if (verbose) {
                    CINFO("planning") << "Solutions not improving anymore.";
                }
                return result;
            }
        }
        
        // apply current action
        if (current_action_opt) { 
            if (current_action_opt->is_special()) {
                if (verbose) {
                    CINFO("planning") << "Special action planned.";
                }
                return result;
            }
            bool success; // not used
            float reward; // not used
            updated_scenario.apply_action(*current_action_opt, success, reward, force_effect);
        }
        else { // no solution -> finished
            if (verbose) {
                CINFO("planning") << "No more solutions.";
            }
            return result;
        }

        // check goal
        if (updated_scenario.is_in_goal()) {
            if (verbose)
                CINFO("planning") << "Reached a goal!";
            reached_goal = true;
            result.push_back(*current_action_opt);
            return result;
        }

        previous_reward = current_reward;
        previous_action_opt = current_action_opt;
    }

    if (verbose) {
        if (result.empty()) {
            CINFO("planning") << " No plan";
        }
        else {
            CINFO("planning") << " Plan is: " << result;
        }
    }

    return result;
}


SymbolOptional IppcPlanner::plan(RexdScenario rexd_scenario, float& planning_result, const bool thoroughly, const bool verbose) const
{
    SymbolOptional resulting_plan;
    rexd_scenario.write_IPPC_PPDDL_to_file(ppddl_template_file, ppddl_rules_file);

    if (verbose) {
        if (thoroughly)
            CINFO("planning") << "Planning full... " << NOENDL;
        else
            CINFO("planning") << "Planning fast... " << NOENDL;
    }

    std::streambuf *cout_buf;
    if (!verbose) {
        cout_buf = stream_quiet(std::cout);
    }
    resulting_plan = ippcPlan(rexd_scenario, thoroughly, planning_result);
    if (!verbose) {
        stream_restore(std::cout, cout_buf);
    }

    if (verbose) {
        if (resulting_plan) {
            CINFO("planning") << " Action: " << START_COLOR_BLUE << *resulting_plan << END_COLOR << " (" << planning_result << ")";
        }
        else {
            CINFO("planning") << " No plan";
        }
    }
    return resulting_plan;
}

void IppcPlanner::execute_glutton_planner_command(const bool thoroughly) const
{
    std::string miliseconds_str;
    //     c++0x
    //     std::string s = std::to_string(42);
    if (thoroughly) {
        miliseconds_str = static_cast<std::ostringstream*>( &(std::ostringstream() << planning_miliseconds) )->str();
    }
    else {
        miliseconds_str = static_cast<std::ostringstream*>( &(std::ostringstream() << planning_fast_miliseconds) )->str();
    }
    std::string epsilon_str = static_cast<std::ostringstream*>( &(std::ostringstream() << planning_epsilon) )->str();
    std::string seed_str = static_cast<std::ostringstream*>( &(std::ostringstream() << rand()) )->str();
    std::string port_str = static_cast<std::ostringstream*>( &(std::ostringstream() << planner_port) )->str();

    std::string glutton_planner_path = planner_path;
    std::string glutton_planner_algorithm = planner_algorithm;
    std::string server = "localhost:" + port_str;
    std::string problem_instance = "ppddl_problem";

    char* glutton_planner_c = utils::string_to_char(glutton_planner_path);
    char* glutton_planner_algorithm_c = utils::string_to_char(glutton_planner_algorithm);
    char* miliseconds_str_c = utils::string_to_char(miliseconds_str);
    char* epsilon_str_c = utils::string_to_char(epsilon_str);
    char* seed_str_c = utils::string_to_char(seed_str);
    char* server_c = utils::string_to_char(server);
    char* ppddl_rules_file_c = utils::string_to_char(ppddl_rules_file);
    char* problem_instance_c = utils::string_to_char(problem_instance);
    char t_param_c[] = "-t";
    char r_param_c[] = "-r";
    char e_param_c[] = "-e";
    char p_param_c[] = "-p";

    char* args[] = { glutton_planner_c, t_param_c, miliseconds_str_c, e_param_c, epsilon_str_c, p_param_c, glutton_planner_algorithm_c, r_param_c, seed_str_c, server_c, ppddl_rules_file_c, problem_instance_c, NULL };

    // silence the planner
    int tmpFd = open( "/dev/null", O_WRONLY );
    dup2( tmpFd, 1 );
    close( tmpFd );

    // wait for server to be ready
    usleep(1000*10); // 10 ms

    // Now execute it
    // TODO signal error if plan doesn't exist
    int res = execvp( args[ 0 ], args );
    if (res == -1)
        printf("ERROR executing execvp. child's execvp: %s\n", strerror(errno));
}


SymbolOptional IppcPlanner::ippcPlan(const RexdScenario& rexd_scenario, const bool thoroughly, float& planning_result) const
{
//     std::string command = get_planner_command(thoroughly);
//     CINFO("planning") << "Planning command is: " << command; // TODO verbose output
//     boost::thread planner_thread(launch_planner, command);
    int pid = fork();
    if ( pid == -1 ) {
        perror("Cannot proceed. fork() error");
        exit(-1);
    }
    if (pid == 0)
    {
        // child process
        utils::initialize_random_seed(); // regenerate rand in the child process
        execute_glutton_planner_command(thoroughly);
        exit(0);
    }

    XMLServer_t *ippcserver;
    ippcserver = new XMLServer_t();
    ippcserver->start_session(planner_port);
    ippcserver->start_round();

    Symbol action = ippcserver->get_action(rexd_scenario.get_state(), planning_result);

    ippcserver->end_round();
    ippcserver->end_session();
    delete ippcserver;

    // wait for server to finish. Not sure if really needed as join/wait would wait for the thread anyway...
//     usleep(1000*10);
    int status;
    wait(&status);
//     planner_thread.join();

    // min result threshold
//     if (planning_result < 1.0) {
//         action = Symbol("");
//     }

    if (utils::trim(action.name) == "") {
        planning_result = -100000.0;
        return SymbolOptional();
    }

    return SymbolOptional(action);
}

