#include "symbolic/domain.h"


/**
 * OPERATORS
 *
 */

std::ostream& operator<<(std::ostream &out, const Domain& dom)
{
    out << "Domain::" << "<<" << " - " << "Domain is :\n";
    for (size_t i = 0; i < dom.known_predicates.size(); i++)
        out << "Domain::" << "<<" << " - " << "Predicate<Symbol> " << i << ": " << dom.known_predicates[i] << "\n";

    return out;
}

/**
 * Domain
 *
 */

Domain::Domain()
{ }

template<class T>
int Domain::index_atom(const T& atom, std::vector<T>& known_atoms)
{
    int idx;
    bool found = false;
    for (size_t i = 0; i < known_atoms.size(); i++) {
        if (atom == known_atoms[i]) {
            idx = i;
            found = true;
        }
    }
    if (!found) {
        idx = known_atoms.size();
        known_atoms.push_back(atom);
    }
    return idx;
}

template<class T>
int Domain::get_atom_index(const T& atom, const std::vector<T>& known_atoms) const
{
    for (size_t i = 0; i < known_atoms.size(); i++) {
        if (atom == known_atoms[i]) {
            return i;
        }
    }
    return -1;
}

void Domain::index_state(const State& state)
{
    index_predicates(state);
    index_objects(state.pddl_object_manager.get_ppddl_objects_per_type());
}

void Domain::index_objects(const PDDLTypeObjectMap& ppddl_objects_per_type)
{
    BOOST_FOREACH(const PDDLTypeObjectMap::value_type& type_objects, ppddl_objects_per_type) {
        BOOST_FOREACH(const std::string& object, type_objects.second) {
            index_atom<std::string>(object, known_objects);
        }
    }
}

void Domain::index_predicates(const PredicateGroup<Symbol>& predicates)
{
    for (PredicateGroup<Symbol>::const_iterator predicate = predicates.begin(); predicate != predicates.end(); ++predicate) {
        index_predicate(*predicate);
    }
}

void Domain::index_goals(const GoalGroup& goals)
{
    for (GoalGroup::const_iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
        index_predicates(goal_it->get_grounded_predicates());
    }
}

void Domain::index_transitions(const TransitionGroup& transitions)
{
    for (TransitionGroup::const_iterator trans_it = transitions.begin(); trans_it != transitions.end(); ++trans_it) {
        index_predicates(trans_it->get_prev_state());
        index_predicates(trans_it->get_post_state());
        index_action(trans_it->action);
    }
}

int Domain::index_predicate(const Predicate<Symbol>& predicate)
{
    Predicate<Symbol> sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = index_atom<Predicate<Symbol> >(predicate, known_predicates);

    // add objects if GroundedPredicate
    if (!predicate.is_symbolic) {
        for (std::vector<std::string>::const_iterator param_it = predicate.param_names.begin(); param_it != predicate.param_names.end(); ++param_it) {
            index_atom<std::string>(*param_it, known_objects);
        }
    }
    return id;
}

int Domain::index_action(const Symbol& action)
{
    Symbol generic_action(action);
    generic_action.make_symbolic();
    return index_atom<Symbol>(generic_action, known_actions);
}

void Domain::index_rules(const RuleSet& rules)
{
    for (RuleList::const_iterator rule_it = rules.begin(); rule_it != rules.end(); ++rule_it) {
        index_rule(**rule_it);
    }
}

void Domain::index_rule(const Rule& rule)
{
    index_action(rule);
    index_predicates(rule.preconditions);
    BOOST_FOREACH(const Outcome& out_it, rule.outcomes) {
        index_predicates(out_it);
    }
}


TransitionGroup Domain::get_transitions_with_indexed_objects(const TransitionGroup& transitions)
{
    TransitionGroup indexed_transitions;
    BOOST_FOREACH(Transition transition, transitions) {
        Transition indexed_transition = get_transition_with_indexed_objects(transition);
        indexed_transitions.push_back(indexed_transition);
    }
    return indexed_transitions;
}

Transition Domain::get_transition_with_indexed_objects(const Transition& transition)
{
    return Transition(get_predicates_with_indexed_objects(transition.prev_state),
                      get_typed_symbol_with_indexed_objects(transition.action),
                      get_predicates_with_indexed_objects(transition.next_state),
                      transition.reward);
}

PredicateGroup<Symbol> Domain::get_predicates_with_indexed_objects(const PredicateGroup<Symbol>& predicates)
{
    PredicateGroup<Symbol> indexed_predicates;
    BOOST_FOREACH(const Predicate<Symbol>& predicate, predicates) {
        indexed_predicates.push_back(get_predicate_with_indexed_objects(predicate));
    }
    return indexed_predicates;
}

Predicate<Symbol> Domain::get_predicate_with_indexed_objects(const Predicate<Symbol>& predicate)
{
    return Predicate<Symbol>(get_symbol_with_indexed_objects(predicate), predicate.positive);
}

Symbol Domain::get_symbol_with_indexed_objects(const Symbol& symbol)
{
    Symbol indexed_symbol(symbol);
    BOOST_FOREACH(std::string& param, indexed_symbol.param_names) {
        // TODO UGLY HACK !!!
        param = boost::lexical_cast<std::string>(get_object_index(param) + 60);
    }
    return indexed_symbol;
}

TypedSymbol Domain::get_typed_symbol_with_indexed_objects(const TypedSymbol& symbol)
{
    TypedSymbol indexed_symbol(symbol);
    BOOST_FOREACH(std::string& param, indexed_symbol.param_names) {
        // TODO UGLY HACK !!!
        param = boost::lexical_cast<std::string>(get_object_index(param) + 60);
    }
    return indexed_symbol;
}

size_t Domain::get_predicate_index(const Predicate<Symbol>& predicate)
{
    Predicate<Symbol> sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = get_atom_index(predicate, known_predicates);
    if (id < 0) { // not found
        id = index_predicate(predicate);
    }
    return id;
}

size_t Domain::get_predicate_index_const(const Predicate<Symbol>& predicate) const
{
    Predicate<Symbol> sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = get_atom_index(predicate, known_predicates);
    return id;
}

size_t Domain::get_action_index(const Symbol& action)
{
    Symbol generic_action(action);
    int id = get_atom_index(generic_action, known_actions);
    if (id < 0) { // not found
        id = index_action(action);
    }
    return id;
}

size_t Domain::get_object_index(const std::string& object)
{
    return get_atom_index(object, known_objects);
}

void Domain::write_PPDDL_grounded_predicates(std::ostream& out) const
{
    std::set<Symbol> predicate_symbol_set; // to write domain we don't need sign(+/-). Predicate<Symbol> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<Symbol> & predicate, known_predicates) {
        predicate_symbol_set.insert(predicate);
    }

    out << "  (:predicates " << std::endl;
    BOOST_FOREACH(const Symbol & symbol, predicate_symbol_set) {
        out << "    ";
        symbol.write_PPDDL(out);
        out << std::endl;
    }
    out << "  )" << std::endl;
}

void Domain::write_PPDDL_symbolic_predicates(std::ostream& out, const PDDLObjectManager& pddl_object_manager) const
{
    // TODO Load actual predicates
    std::set<TypedSymbol> typed_symbol_set; // to write domain we don't need sign(+/-). Predicate<Symbol> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<Symbol> & predicate, known_predicates) {
        PDDLTypeVector pddl_type_vector;
        BOOST_FOREACH(const std::string& object, predicate.param_names) {
            pddl_type_vector.push_back(pddl_object_manager.get_object_type(object));
        }
        TypedSymbol typed_symbol(predicate.name, predicate.param_names, pddl_type_vector);
        typed_symbol.make_symbolic();
        if ( (typed_symbol.param_types.empty()) || (typed_symbol.param_types.front() != OBJECT_TYPE_DEFAULT) ) {
            // duplicates predicates because we don't know type for ?X or ?Y in some symbolic predicates
            typed_symbol_set.insert(typed_symbol);
        }
    }

    out << "  (:predicates ";
    BOOST_FOREACH(const TypedSymbol & typed_symbol, typed_symbol_set) {
        out << "               ";
        typed_symbol.write_PPDDL(out);
        out << std::endl;
    }
    out << "  )" << std::endl;
}

void Domain::write_NID(std::ostream& out) const
{
    std::set<Symbol> predicate_symbol_set, action_symbol_set; // to write domain we don't need sign(+/-). Predicate<Symbol> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<Symbol> & predicate, known_predicates) {
        predicate_symbol_set.insert(predicate);
    }

    BOOST_FOREACH(const Symbol & action, known_actions) {
        action_symbol_set.insert(action);
    }

    out << "# predicates" << std::endl;
    BOOST_FOREACH(const Symbol & primitive, predicate_symbol_set) {
        out << primitive.name << " " << primitive.param_names.size() << " primitive binary" << std::endl;
    }

    out << "# actions" << std::endl;
    BOOST_FOREACH(const Symbol & action, action_symbol_set) {
        out << action.name << " " << action.param_names.size() << " action binary" << std::endl;
    }
}





