#include <symbolic/goals.h>


std::ostream& operator<<(std::ostream &out, const Goal& g)
{
    out << g.reward << " ";
    out << g.grounded_goals;
    out << " ";
    out << g.typed_symbolic_goals;
    out << std::flush;
    return out;
}

std::ostream& operator<<(std::ostream &out, const GoalGroup& g)
{
    for (GoalGroup::const_iterator goal_it = g.begin(); goal_it != g.end(); ++goal_it) {
        out << *goal_it << std::endl;
    }
    return out;
}

void Goal::write_PPDDL(std::ostream &out) const
{
    assert(typed_symbolic_goals.empty() || grounded_goals.empty()); // either symbolic or grounded, but not both
    if (typed_symbolic_goals.size() > 0) {
        out << "                (when  ";
        typed_symbolic_goals.write_PPDDL(out);
    }
    else if (grounded_goals.size() > 0) {
        out << "                (when  ";
        grounded_goals.write_PPDDL(out);
    }
    out << " (";
    if (reward >= 0)
        out << "increase ";
    else
        out << "decrease ";
    out << "(reward) ";
    out << std::abs(reward);
    out << ")";
    if ( (typed_symbolic_goals.size() > 0) || (grounded_goals.size() > 0) ) {
        out << ")";
    }
    out << std::endl;
}

void GoalGroup::write_PPDDL(std::ostream &out) const
{
//     out << std::endl << "    ; Reward" << std::endl;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        goal_it->write_PPDDL(out);
    }
}


// Goal
Goal::Goal(const std::string& goal_string):
    reward(0)
{
    std::string ss_item;
    std::stringstream goal_ss(goal_string);
    std::getline(goal_ss, ss_item, ' ');
    try {
        reward = boost::lexical_cast<float>(ss_item);
        std::getline(goal_ss, ss_item);
        if (goal_ss.str() == ss_item) { // if no predicates getline does not obtain anything new
            ss_item = "";
        }
        add_predicates(PredicateGroup<Symbol>(ss_item));
    } catch (boost::bad_lexical_cast &) {
        LOG(ERROR) << "Error parsing goal " << goal_string;
        LOG(WARNING) << "Compatibility mode, setting reward to 1.0 and goal predicates to " << goal_string;
        reward = 1.0;
        add_predicates(PredicateGroup<Symbol>(goal_string));
    }
}

float Goal::obtain_reward(const PredicateGroup< Symbol >& predicates) const
{
    float obtained_reward = 0;
    if (grounded_goals.is_satisfied_by(predicates)) {
        obtained_reward += this->reward;
    }
    return obtained_reward;
}

GoalGroup Goal::get_pddl_grounded_goals(const std::vector< std::string >& symbolic_params, 
                                        const std::vector< std::string >& grounded_params, 
                                        const PDDLObjectManager& object_manager) const
{
    GoalGroup result_goals;
    
    Goal copy_goal(*this);
    copy_goal.typed_symbolic_goals.ground_params(symbolic_params, grounded_params);
    
    // get symbolic params with their corresponding type
    PDDLObjectTypeMap symbolic_object_type_params;
    BOOST_FOREACH(const Predicate<TypedSymbol>& goal_pred, copy_goal.typed_symbolic_goals) {
        for (size_t i = 0; i < goal_pred.param_types.size(); i++) {
            if ( symbolic::is_param_symbolic(goal_pred.param_names[i]) ) {
                symbolic_object_type_params[goal_pred.param_names[i]] = goal_pred.param_types[i];
            }
        }
    }
    
    // swap typed_symbolic_goals to grounded_goals
    BOOST_FOREACH(Predicate<TypedSymbol> typed_pred, copy_goal.typed_symbolic_goals) {
        copy_goal.grounded_goals.push_back(Predicate<Symbol>(typed_pred, typed_pred.positive));
    }
    copy_goal.typed_symbolic_goals.clear();
    result_goals.push_back(copy_goal);
    
    // create all groundings for each param based on its types
    BOOST_FOREACH(const PDDLObjectTypeMap::value_type& object_type_param, symbolic_object_type_params) {
        GoalGroup new_goals;
        BOOST_FOREACH(const Goal& result_goal, result_goals) {
            BOOST_FOREACH(const std::string& object, object_manager.get_type_objects(object_type_param.second)) {
                Goal new_goal(result_goal);
                new_goal.grounded_goals.change_variable(object_type_param.first, object);
                new_goals.push_back(new_goal);
            }
        }
        result_goals = new_goals;
    }
    
//     LOG(INFO) << "Grounded goals are " << result_goals;
    return result_goals;
}


// GoalGroup
GoalGroup::GoalGroup(std::string file_path)
{
    std::ifstream goals_file(file_path.c_str());
    if (goals_file.is_open()) {
        while (goals_file.good()) {
            std::string goal_string;
            std::getline(goals_file, goal_string);
            goal_string = utils::trim(goal_string);
            if (goal_string != "") {
                push_back(Goal(goal_string));
            }
        }
        goals_file.close();
    } else LOG(ERROR) << "Unable to open file: "  << file_path;
}

bool GoalGroup::is_goal(const PredicateGroup<Symbol>& state) const
{
    bool result = true;
    if (!state.has_predicate(Predicate<Symbol>(MAKE_SYMBOL_FINISHED, true))) { // if not finished
        for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) { // check if all goals are satisfied
            if (!goal_it->is_goal(state)) {
                result = false;
            }
        }
    }
    return result;
}


PredicateGroup<Symbol> GoalGroup::get_predicates() const
{
    PredicateGroup<Symbol> predicates;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        predicates.add_predicates(goal_it->get_grounded_predicates());
    }
    return predicates;
}

float GoalGroup::obtain_reward(const PredicateGroup<Symbol>& predicates) const
{
    float obtained_reward = 0;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        obtained_reward += goal_it->obtain_reward(predicates);
    }
    return obtained_reward;
}

Goal GoalGroup::adapt_finish_rule_goal_to_simple_predicate_goal()
{
    Goal simple_goal(1.0);

    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        PredicateGroup<Symbol> goal_predicates = goal_it->get_grounded_predicates();
        goal_predicates.remove(Predicate<Symbol>(MAKE_SYMBOL_FINISHED, false));
        if (goal_it->get_reward() < 0.0) {
            goal_predicates = goal_predicates.get_negatet_predicates();
        }
        simple_goal.add_predicates(goal_predicates);
    }

    return simple_goal;
}

GoalGroup GoalGroup::get_pddl_grounded_goals(const std::vector< std::string >& symbolic_params, const std::vector< std::string >& grounded_params, const PDDLObjectManager& pddl_object_manager)
{
    GoalGroup result;
    for (GoalGroup::iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        std::vector<Goal> new_grounded_goals = goal_it->get_pddl_grounded_goals(symbolic_params, grounded_params, pddl_object_manager);
        result.insert(result.end(), new_grounded_goals.begin(), new_grounded_goals.end());
    }
    return result;
}

void GoalGroup::PPDDL_grounding()
{
    for (GoalGroup::iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        goal_it->PPDDL_grounding();
    }
}

void GoalGroup::transform_symbolic_NID_to_symbolic_PDDL()
{
    BOOST_FOREACH(Goal& goal, *this) {
        goal.transform_symbolic_NID_to_symbolic_PDDL();
    }
}

