/*
 * Clases representing predicates and goals.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_GOALS_H
#define IRI_RULE_LEARNER_GOALS_H 1


#include <boost/lexical_cast.hpp>

#include <symbolic/predicates.h>


class GoalGroup;

/*!
\class Goal
\brief A vector of predicates with an associated reward.

The reward is obtained when the set of predicates are satisfied.
*/
class Goal
{
private:
    float reward;
    PredicateGroup<TypedSymbol> typed_symbolic_goals;
    PredicateGroup<Symbol> grounded_goals;
    
public:
    /**
     * Constructors
     */
    Goal(float reward_ = 1.0) :
        reward(reward_)
    { }

    Goal(PredicateGroup<Symbol> grounded_predicates_, float reward_ = 1.0) :
        reward(reward_),
        grounded_goals(grounded_predicates_)
    { }
    
    Goal(PredicateGroup<TypedSymbol> typed_predicates_, float reward_ = 1.0) :
        reward(reward_),
        typed_symbolic_goals(typed_predicates_)
    { }
    
    Goal(const Goal& other):
        reward(other.reward),
        typed_symbolic_goals(other.typed_symbolic_goals),
        grounded_goals(other.grounded_goals)
    { }

    Goal(const std::string& goal_string);
    
    bool is_typed() { return ( (grounded_goals.empty()) && (typed_symbolic_goals.size() > 0) ); }
    
    void add_predicates(const PredicateGroup<Symbol>& predicates)      { assert(typed_symbolic_goals.empty()); grounded_goals.add_predicates(predicates); };
    void add_predicates(const PredicateGroup<TypedSymbol>& predicates) { assert(grounded_goals.empty()); typed_symbolic_goals.add_predicates(predicates); };

    /** \brief true state satisfies the PredicateGroup */
    bool is_goal(const PredicateGroup<Symbol> state) const {
        return grounded_goals.is_satisfied_by(state);
    };
    
    /** \brief returns the goal reward  if predicates satisfies the PredicateGroup */
    float obtain_reward(const PredicateGroup<Symbol>& predicates) const;
    
    /** \brief return all grounded predicates in the goal */
    const PredicateGroup<Symbol>& get_grounded_predicates() const { return grounded_goals; }
    
    /** \brief get reward */
    float get_reward() const { return reward; }
    
    /** \brief set reward */
    void set_reward(const float new_reward) { reward = new_reward; }
    
    /** \brief do ppddl grounding */
    void PPDDL_grounding() { grounded_goals.PPDDL_grounding(); typed_symbolic_goals.PPDDL_grounding(); }
    
    /** \brief transform_symbolic_NID_to_symbolic_PDDL */
    void transform_symbolic_NID_to_symbolic_PDDL() { grounded_goals.transform_symbolic_NID_to_symbolic_PDDL(); typed_symbolic_goals.transform_symbolic_NID_to_symbolic_PDDL(); };
    
    /**
     * \brief returns all grounded goals
     * 
     * symbolic_params are grounded with grounded_params, while all other params are grounded to each of in the list_of_objects
     */
    GoalGroup get_pddl_grounded_goals(const std::vector< std::string >& symbolic_params, 
                                      const std::vector< std::string >& grounded_params, 
                                      const PDDLObjectManager& object_manager) const;

    void write_PPDDL(std::ostream &out) const;

    /**
     * Operators
     */
    friend std::ostream& operator<<(std::ostream &out, const Goal& g);

    // TODO goal operators
//         friend bool operator==(const Goal &o1, const Goal &o2);
};


/*!
\class GoalGroup
\brief A vector of goals
*/
class GoalGroup : public std::vector<Goal>
{
public:
    GoalGroup() {}
    GoalGroup(std::string file_path);

    void PPDDL_grounding();
    
    /**
     * \brief returns all grounded goals
     * 
     * symbolic_params are grounded with grounded_params, while all other params are grounded to each of in the list_of_objects
     */
    GoalGroup get_pddl_grounded_goals(const std::vector< std::string >& symbolic_params, const std::vector< std::string >& grounded_params, const PDDLObjectManager& pddl_object_manager);

    bool is_goal(const PredicateGroup<Symbol>& state) const;
    float obtain_reward(const PredicateGroup<Symbol>& predicates) const;

    /** \brief Adapts a goal defined by a finish function to a goal defined by a set of predicates giving a positive reward when satisfied
     *
     * DEPRECATED
     * Just used for backward compatibility.
     * Gets a set of goals which are the rewards obtained by a finish function.\n
     *  - Ex: [+1.5 onTable(1), -1.0 free(2)]
     * \return A set of predicates which give positive rewards
     */
    Goal adapt_finish_rule_goal_to_simple_predicate_goal();

    /** \brief Transforms X to ?X, Y to ?Y, etc... */
    void transform_symbolic_NID_to_symbolic_PDDL();
    
    /** \brief Get vector with all predicates in all goals */
    PredicateGroup<Symbol> get_predicates() const;

    void write_PPDDL(std::ostream &out) const;

    friend std::ostream& operator<<(std::ostream &out, const GoalGroup& g);
};

typedef boost::shared_ptr<Goal> GoalPtr;
typedef boost::shared_ptr<const Goal> GoalConstPtr;

typedef boost::shared_ptr<GoalGroup> GoalGroupPtr;
typedef boost::shared_ptr<const GoalGroup> GoalGroupConstPtr;



#endif

