/*
 * <one line to give the program's name and a brief idea of what it does.>
 * Copyright (C) 2014  dmartinez <email>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pddl_parser.h"

#define PDDL_KEY_DEFINE       "define"
#define PDDL_KEY_PROBLEM      ":problem"
#define PDDL_KEY_DOMAIN       ":domain"
#define PDDL_KEY_REQUIREMENTS ":requirements"
#define PDDL_KEY_TYPES        ":types"
#define PDDL_KEY_PREDICATES   ":predicates"
#define PDDL_KEY_ACTION       ":action"
#define PDDL_KEY_OBJECTS      ":objects"
#define PDDL_KEY_INIT         ":init"
#define PDDL_KEY_GOAL         ":goal"
#define PDDL_KEY_GOAL_REWARD  ":goal-reward"
#define PDDL_KEY_METRIC       ":metric"

#define PDDL_ELEMENT_NEGATION      "not"
#define PDDL_ELEMENT_AND           "and"
#define PDDL_ELEMENT_PROBABILISTIC "probabilistic"
#define PDDL_ELEMENT_WHEN          "when"
#define PDDL_ELEMENT_INCREASE      "increase"
#define PDDL_ELEMENT_DECREASE      "decrease"

#define PDDL_ACTION_IDENTIFIER   ":action"
#define PDDL_ACTION_PARAMETERS   ":parameters"
#define PDDL_ACTION_PRECONDITION ":precondition"
#define PDDL_ACTION_EFFECT       ":effect"


PDDLParser::action_string_code PDDLParser::enum_hash (std::string const& str) const {
    if (str == PDDL_ACTION_PARAMETERS) 
        return ePDDL_ACTION_PARAMETERS;
    else if (str == PDDL_ACTION_PRECONDITION) 
        return ePDDL_ACTION_PRECONDITION;
    else if (str == PDDL_ACTION_EFFECT) 
        return ePDDL_ACTION_EFFECT;
    else if (str == PDDL_KEY_DEFINE) 
        return ePDDL_KEY_DEFINE;
    else if (str == PDDL_KEY_DOMAIN) 
        return ePDDL_KEY_DOMAIN;
    else if (str == PDDL_KEY_GOAL) 
        return ePDDL_KEY_GOAL;
    else if (str == PDDL_KEY_GOAL_REWARD) 
        return ePDDL_KEY_GOAL_REWARD;
    else if (str == PDDL_KEY_INIT) 
        return ePDDL_KEY_INIT;
    else if (str == PDDL_KEY_METRIC) 
        return ePDDL_KEY_METRIC;
    else if (str == PDDL_KEY_OBJECTS) 
        return ePDDL_KEY_OBJECTS;
    else if (str == PDDL_KEY_PREDICATES) 
        return ePDDL_KEY_PREDICATES;
    else if (str == PDDL_KEY_PROBLEM) 
        return ePDDL_KEY_PROBLEM;
    else if (str == PDDL_KEY_REQUIREMENTS) 
        return ePDDL_KEY_REQUIREMENTS;
    else if (str == PDDL_KEY_TYPES) 
        return ePDDL_KEY_TYPES;
    else
        return ePDDL_UNKNOWN;
}



namespace fusion = boost::fusion;
namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;


PDDLParser::PDDLParser()
{

}


template <typename Iterator, typename Skipper = qi::rule<Iterator> >
    struct pddl_grammar : qi::grammar<Iterator, PDDLReverseStruct(), Skipper>
    {
        pddl_grammar() : pddl_grammar::base_type(pddl_struct)
        {
            using qi::lit;
            using qi::lexeme;
            using ascii::char_;
            using ascii::string;
            using namespace qi::labels;

            using phoenix::at_c;
            using phoenix::push_back;
            
            text %= lexeme[+(char_ - '(' - ')' - ' ' - '\n')];
            node %= (pddl_struct | text);

            start_tag = lit('(');
            end_tag = lit(')');

            pddl_struct =
                    start_tag                   
                >>  *(node                       [push_back(at_c<0>(_val), _1)])
                >>  end_tag
            ;
        }

        qi::rule<Iterator, PDDLReverseStruct(), Skipper> pddl_struct;
        qi::rule<Iterator, PDDLElement(), Skipper> node;
        qi::rule<Iterator, std::string(), Skipper> text;
        qi::rule<Iterator, std::string(), Skipper> start_tag;
        qi::rule<Iterator, std::string(), Skipper> end_tag;
    };


PDDLReverseStruct PDDLParser::read_pddl_reverse(const std::string& str) const
{
    using boost::spirit::ascii::space; 
    
    // Comment skipper
    qi::rule<std::string::const_iterator> skip_ws_and_comments 
            = space 
            | ";" >> *(ascii::char_-qi::eol) >> qi::eol
            ;
    
    PDDLReverseStruct pddl_reverse_struct;
    pddl_grammar<std::string::const_iterator, qi::rule<std::string::const_iterator> > g; // grammar
    bool r = phrase_parse(str.begin(), str.end(), g, skip_ws_and_comments, pddl_reverse_struct);
    if (!r) {
        LOG(WARNING) << "There was an error parsing " << str;
    }
    return pddl_reverse_struct;
}

template<>
Symbol PDDLParser::parse_symbol<Symbol> (const PDDLReverseStruct& pddl_reverse_struct) const
{
    std::string name;
    std::vector<std::string> param_names;
    
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    name = boost::get<std::string>(*elem_it); // get first element
    
    for(++elem_it; elem_it != pddl_reverse_struct.children.end(); ++elem_it) {
        if (boost::get<std::string>(*elem_it) == "-") { // was not a properly defined Symbol
            throw PDDLException("Symbol");
        }   
        param_names.push_back(boost::get<std::string>(*elem_it));
    }
    
    Symbol symbol(name, param_names);
    
//     LOG(DEBUG) << "Symbol is " << symbol;
    
    return symbol;
}

std::vector<std::string> PDDLParser::parse_string_list (const PDDLReverseStruct& pddl_reverse_struct) const
{
    return parse_string_list(pddl_reverse_struct, pddl_reverse_struct.children.begin());
}

std::vector<std::string> PDDLParser::parse_string_list (const PDDLReverseStruct& pddl_reverse_struct, std::vector<PDDLElement>::const_iterator elem_it) const
{
    std::vector<std::string> result;
    for(; elem_it != pddl_reverse_struct.children.end(); ++elem_it) {
        result.push_back(boost::get<std::string>(*elem_it));
    }
    return result;
}

void PDDLParser::parse_typed_objects (const PDDLReverseStruct& pddl_reverse_struct, std::vector<PDDLElement>::const_iterator elem_it, 
                                                  std::vector<std::string>& object_names, std::vector<std::string>& object_types) const
{
    std::vector<std::string> current_objects;
    while (elem_it != pddl_reverse_struct.children.end()) {
        std::string current_elem = boost::get<std::string>(*elem_it);
        if (current_elem != "-") {
            current_objects.push_back(current_elem);
            ++elem_it;
        }
        else {
            std::string  current_type = boost::get<std::string>(*(elem_it+1));
            
            BOOST_FOREACH(std::string object, current_objects) {
                object_names.push_back(object);
                object_types.push_back(current_type);
            }
            
            current_objects.clear();
            elem_it = elem_it + 2;
        }
    }
    if (!current_objects.empty()) { // was not a properly defined TypedSymbol
        throw PDDLException("TypedSymbol");
    }
}

template<>
TypedSymbol PDDLParser::parse_symbol<TypedSymbol> (const PDDLReverseStruct& pddl_reverse_struct) const
{
    std::string name;
    std::vector<std::string> param_names;
    std::vector<std::string> param_types;
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    name = boost::get<std::string>(*elem_it); // get first element
    parse_typed_objects(pddl_reverse_struct, ++elem_it, param_names, param_types);
    
    TypedSymbol typed_symbol(name, param_names, param_types);
    
//     LOG(DEBUG) << "TypedSymbol is " << typed_symbol;
    
    return typed_symbol;
}

template<typename SymbolT>
Predicate<SymbolT>   PDDLParser::parse_predicate       (const PDDLReverseStruct& pddl_reverse_struct) const
{
    bool positive;
    SymbolT symbol_t;
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    
    if (boost::get<std::string>(*elem_it) == PDDL_ELEMENT_NEGATION) { // (not (pred1))
        positive = false;
        symbol_t = parse_symbol<SymbolT>(boost::get<PDDLReverseStruct>(*(elem_it+1)));
    }
    else { // (pred1)
        positive = true;
        symbol_t = parse_symbol<SymbolT>(boost::get<PDDLReverseStruct>(pddl_reverse_struct));
    }
    Predicate<SymbolT> predicate(symbol_t, positive);
    
//     LOG(DEBUG) << "Predicate<Symbol> is " << predicate;
    return predicate;
}

template<typename SymbolT>
PredicateGroup<SymbolT> PDDLParser::parse_predicate_group (const PDDLReverseStruct& pddl_reverse_struct) const
{
    PredicateGroup<SymbolT> predicates;
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    std::string first_element = boost::get<std::string>(*elem_it);
    if ( (first_element == PDDL_ELEMENT_AND) || // (and (pred1) (pred2) ... )
         (first_element == PDDL_KEY_INIT) )     // (:init (pred1) (pred2) ... )
    {
        for(++elem_it; elem_it != pddl_reverse_struct.children.end(); ++elem_it) {
            predicates.push_back(parse_predicate<SymbolT>(boost::get<PDDLReverseStruct>(*elem_it)));
        }
    }
    else { // (pred1)
        predicates.push_back(parse_predicate<SymbolT>(pddl_reverse_struct));
    }
    
//     LOG(DEBUG) << "Predicates is " << predicates;
    return predicates;
}

OutcomeGroup     PDDLParser::parse_outcome         (const PDDLReverseStruct& pddl_reverse_struct) const
{
    OutcomeGroup outcomes;
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    if (boost::get<std::string>(*elem_it) == PDDL_ELEMENT_WHEN) { // (when ...)
        // ignoring
    }
    else if (boost::get<std::string>(*elem_it) == PDDL_ELEMENT_PROBABILISTIC) { // (probabilistic 0.5 (pred_group))
        for (++elem_it; elem_it != pddl_reverse_struct.children.end(); ++elem_it) {
            float probability = boost::lexical_cast<float>(boost::get<std::string>(*elem_it));
            elem_it++;
            PredicateGroup<Symbol> predicates = parse_predicate_group<Symbol>(boost::get<PDDLReverseStruct>(*elem_it));
            outcomes.push_back(Outcome(predicates, probability));
        }
        outcomes.add_empty_outcome_to_fill_probability();
    }
    else { // (pred1)
        outcomes.push_back(Outcome(parse_predicate_group<Symbol>(pddl_reverse_struct), 1.0));
    }
    
//     LOG(DEBUG) << "PPDDL outcome is " << outcomes;
    
    return outcomes;
}

OutcomeGroup   PDDLParser::parse_outcome_group   (const PDDLReverseStruct& pddl_reverse_struct) const
{
    OutcomeGroup nid_outcomes;
    nid_outcomes.push_back(Outcome(1.0)); // add empty outcome with prob = 1.0 to start
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    if (boost::get<std::string>(*elem_it) == PDDL_ELEMENT_AND) { // (and (out1) (out2))
        for(++elem_it; elem_it != pddl_reverse_struct.children.end(); ++elem_it) {
            OutcomeGroup pddl_outcomes = parse_outcome(boost::get<PDDLReverseStruct>(*elem_it));
            // assert(sum of probabilities == 1);
            if (pddl_outcomes.size() == 1) {
                BOOST_FOREACH(Outcome& nid_outcome, nid_outcomes) {
                    nid_outcome.add_predicates(pddl_outcomes.front());
                }
            }
            else if (pddl_outcomes.size() > 1) {
                OutcomeGroup new_nid_outcomes;
                BOOST_FOREACH(const Outcome& pddl_outcome, pddl_outcomes) {
                    OutcomeGroup copy_nid_outcomes(nid_outcomes);
                    BOOST_FOREACH(Outcome& nid_outcome, copy_nid_outcomes) {
                        nid_outcome.probability *= pddl_outcome.probability;
                        nid_outcome.add_predicates(pddl_outcome);
                    }
                    new_nid_outcomes.add_new_outcomes(copy_nid_outcomes);
                }
                nid_outcomes = new_nid_outcomes;
            }
        }
    }
    else { // (out1)
        OutcomeGroup outcomes = parse_outcome(pddl_reverse_struct);
        nid_outcomes.add_new_outcomes(outcomes);
    }
    
    nid_outcomes.reorder_by_probability();
    
//     LOG(DEBUG) << "Outcomes are " << nid_outcomes;
    return nid_outcomes;
}

boost::optional<Goal> PDDLParser::parse_goal (const PDDLReverseStruct& pddl_reverse_struct) const
{
    boost::optional<Goal> goal_opt;
    
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    if (boost::get<std::string>(*elem_it) == PDDL_ELEMENT_WHEN) { // (when (pred_group) (decrease/increase (reward) 1.0))
        elem_it++;
        PredicateGroup<TypedSymbol> typed_predicates;
        PredicateGroup<Symbol> grounded_predicates;
        try {
            grounded_predicates = parse_predicate_group<Symbol>(boost::get<PDDLReverseStruct>(*elem_it));
        }
        catch (PDDLException& e) {
            typed_predicates = parse_predicate_group<TypedSymbol>(boost::get<PDDLReverseStruct>(*elem_it));
        }
        elem_it++;
        PDDLReverseStruct reward_pddl_struct = boost::get<PDDLReverseStruct>(*elem_it);
        std::vector<PDDLElement>::const_iterator reward_pddl_struct_it = reward_pddl_struct.children.begin();
        std::string increase_or_decrease = boost::get<std::string>(*reward_pddl_struct_it);
        reward_pddl_struct_it++;
        // assert ("reward")
        std::string reward_string = boost::get<std::string>(boost::get<PDDLReverseStruct>(*reward_pddl_struct_it).children.front());
        assert(reward_string == "reward");
        // reward
        reward_pddl_struct_it++;
        float reward = boost::lexical_cast<float>(boost::get<std::string>(*reward_pddl_struct_it));
        
        if (increase_or_decrease == PDDL_ELEMENT_INCREASE) {
            if (typed_predicates.empty()) {
                goal_opt = Goal(grounded_predicates, reward);
            }
            else {
                goal_opt = Goal(typed_predicates, reward);
            }
        }
        else if (increase_or_decrease == PDDL_ELEMENT_DECREASE) {
            reward *= -1;
            if (typed_predicates.empty()) {
                goal_opt = Goal(grounded_predicates, reward);
            }
            else {
                goal_opt = Goal(typed_predicates, reward);
            }
        }
        else {
            LOG(WARNING) << "Failed to read PDDL goal";
        }
    }
    
    return goal_opt;
}

GoalGroup PDDLParser::parse_goal_group (const PDDLReverseStruct& pddl_reverse_struct) const
{
    GoalGroup goals;
    std::vector<PDDLElement>::const_iterator outcome_it = pddl_reverse_struct.children.begin();
    if (boost::get<std::string>(*outcome_it) == PDDL_ELEMENT_AND) { // (and (out1) (out2) ... (when reward1) (when reward2) ...)
        for(++outcome_it; outcome_it != pddl_reverse_struct.children.end(); ++outcome_it) {
            boost::optional<Goal> goal_opt = parse_goal(boost::get<PDDLReverseStruct>(*outcome_it));
            if (goal_opt) {
                goals.push_back(*goal_opt);
            }
        }
    }
    else { // (out1)
        boost::optional<Goal> goal_opt = parse_goal(pddl_reverse_struct);
        if (goal_opt) {
            goals.push_back(*goal_opt);
        }
    }
    
//     LOG(DEBUG) << "Goals are " << goals;
    return goals;
}

SymbolicRulePtr PDDLParser::parse_symbolic_rule   (const PDDLReverseStruct& pddl_reverse_struct) const
{
    std::string action_name;
    std::vector<std::string> param_names;
    std::vector<std::string> param_types;
    PredicateGroup<Symbol> preconditions;
    OutcomeGroup outcomes;
    GoalGroup goals;
    
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    if (!(boost::get<std::string>(*elem_it) == PDDL_ACTION_IDENTIFIER)) { // (:action action_name ...
        throw PDDLException("SymbolicRule");
    }
    ++elem_it;
    action_name = boost::get<std::string>(*elem_it);
    
    for(++elem_it; elem_it != pddl_reverse_struct.children.end(); elem_it+=2) {
        PDDLReverseStruct params_pddl_struct = boost::get<PDDLReverseStruct>(*(elem_it+1));
        switch (enum_hash(boost::get<std::string>(*elem_it))) {
            case ePDDL_ACTION_PARAMETERS:
                try {
                    parse_typed_objects(params_pddl_struct, params_pddl_struct.children.begin(), param_names, param_types);
                }
                catch (boost::bad_get& e) {
                    LOG(WARNING) << "Error reading rule " << action_name << " parameters: " << e.what();
                }
                break;
            case ePDDL_ACTION_PRECONDITION:
                try {
                    preconditions = parse_predicate_group<Symbol>(params_pddl_struct);
                }
                catch (boost::bad_get& e) {
                    LOG(WARNING) << "Error reading rule " << action_name << " preconditions: " << e.what();
                }
                break;
                
            case ePDDL_ACTION_EFFECT:
                try {
                    outcomes = parse_outcome_group(params_pddl_struct);
                    goals = parse_goal_group(params_pddl_struct);
                }
                catch (boost::bad_get& e) {
                    LOG(WARNING) << "Error reading rule " << action_name << " effects: " << e.what();
                }   
                break;
                
            default:
                LOG(WARNING) << "Didn't recognize action element: " << boost::get<std::string>(*elem_it);
                break;
        }
    }
    
    SymbolicRulePtr resulting_rule(new SymbolicRule(TypedSymbol(action_name, param_names, param_types), preconditions, outcomes, goals));
    
    // add rule preconditions to goals
    BOOST_FOREACH(Goal& goal, resulting_rule->goals) {
        if (goal.is_typed()) {
            goal.add_predicates(resulting_rule->get_typed_preconditions());
        }
        else {
            goal.add_predicates(preconditions);
        }
    }
    
    return resulting_rule;
}

RuleSetPtr        PDDLParser::parse_symbolic_ruleset(const PDDLReverseStruct& pddl_reverse_struct) const
{
    RuleSetPtr rules_ptr = RuleSetPtr(new RuleSet());
    
    BOOST_FOREACH(const PDDLElement& pddl_element, pddl_reverse_struct.children) {
        try {
            PDDLReverseStruct pddl_possible_rule_element;
            try {
                pddl_possible_rule_element = boost::get<PDDLReverseStruct>(pddl_element);
            }
            catch (boost::bad_get& e) {
                // We are using this parser to read the whole domain file
                // Ignoring not PDDLReverseStruct elements. Example: the first "(define" field
                throw PDDLException("Ignoring non-recursive element");
            }
            
            SymbolicRulePtr rule = parse_symbolic_rule(pddl_possible_rule_element);
            rules_ptr->push_back(rule);
        }
        catch (PDDLException& e) {
            // expected as we are passing all types of things, not just actions
//             LOG(ERROR) << "Exception when reading rule set: " << e.what();
        }
    }
    
    return rules_ptr;
}


Symbol PDDLParser::read_symbol          (const std::string& str) const
{
    return parse_symbol<Symbol>(read_pddl_reverse(str));
}

TypedSymbol PDDLParser::read_typed_symbol    (const std::string& str) const
{
    return parse_symbol<TypedSymbol>(read_pddl_reverse(str));
}

Predicate<Symbol> PDDLParser::read_predicate       (const std::string& str) const
{
    return parse_predicate<Symbol>(read_pddl_reverse(str));
}

PredicateGroup<Symbol> PDDLParser::read_predicate_group (const std::string& str) const
{
    return parse_predicate_group<Symbol>(read_pddl_reverse(str));
}

OutcomeGroup PDDLParser::read_outcome_group   (const std::string& str) const
{
    return parse_outcome_group(read_pddl_reverse(str));
}

SymbolicRulePtr PDDLParser::read_symbolic_rule   (const std::string& str) const
{
    return parse_symbolic_rule(read_pddl_reverse(str));
}

RuleSetPtr PDDLParser::read_symbolic_ruleset(const std::string& str) const
{
    return parse_symbolic_ruleset(read_pddl_reverse(str));
}

PDDLReverseStructOpt PDDLParser::find_domain_field(const std::string& pddl_str, const std::string& pddl_key) const
{
    PDDLReverseStruct pddl_reverse_struct = read_pddl_reverse(pddl_str);
    
    std::vector<PDDLElement>::const_iterator elem_it = pddl_reverse_struct.children.begin();
    if (!(boost::get<std::string>(*elem_it) == PDDL_KEY_DEFINE)) { // (define (problem/domain
        throw PDDLException("PDDL file type not defined");
    }
    ++elem_it;
    // ( [problem/domain] [problem_name/domain_name] )
    // boost::get<PDDLReverseStruct>(*elem_it); 
    
    for(++elem_it ; elem_it != pddl_reverse_struct.children.end(); ++elem_it) {
        PDDLReverseStruct pddl_field = boost::get<PDDLReverseStruct>(*elem_it); 
        if (boost::get<std::string>(pddl_field.children.front()) == pddl_key) {
            return PDDLReverseStructOpt(pddl_field);
        }
    }
    return PDDLReverseStructOpt();
}

PDDLObjectManager PDDLParser::parse_pddl_object_list(const PDDLReverseStruct& pddl_reverse_struct) const
{
    PDDLObjectManager pddl_object_manager;
    std::vector<std::string> object_names;
    std::vector<std::string> object_types;
    // skip first element ":objects"
    parse_typed_objects(pddl_reverse_struct, pddl_reverse_struct.children.begin()+1, object_names, object_types);
    
    pddl_object_manager.set_ppddl_objects_from_object_type_pairs(object_names, object_types);
    return pddl_object_manager;
}

PDDLObjectManager PDDLParser::read_objects_from_simple_string(const std::string& str) const
{
    return parse_pddl_object_list(read_pddl_reverse(str));
}

State PDDLParser::read_state(const std::string& str) const
{
    State state;
    
    PDDLReverseStructOpt pddl_reverse_struct_opt = find_domain_field(str, PDDL_KEY_INIT);
    if (pddl_reverse_struct_opt) {
        state = State( parse_predicate_group<Symbol>(*pddl_reverse_struct_opt) );
    }
    
    pddl_reverse_struct_opt = find_domain_field(str, PDDL_KEY_OBJECTS);
    if (pddl_reverse_struct_opt) {
        state.pddl_object_manager = parse_pddl_object_list(*pddl_reverse_struct_opt);
    }
    
    return state;
}


GoalGroup PDDLParser::read_global_goals(const std::string& str) const
{
    GoalGroup goals;
    
    float goal_reward = 1.0;
    PDDLReverseStructOpt pddl_reverse_struct_opt = find_domain_field(str, PDDL_KEY_GOAL_REWARD);
    if (pddl_reverse_struct_opt) {
        // Second element in pddl_reverse_struct_opt is the reward. boost::variant -> std::string -> float
        goal_reward = boost::lexical_cast<float>(boost::get<std::string>(*(pddl_reverse_struct_opt->children.begin()+1)));
    }
    
    pddl_reverse_struct_opt = find_domain_field(str, PDDL_KEY_GOAL);
    if (pddl_reverse_struct_opt) {
        // remove initial pddl key :goal, start in second field
        PDDLReverseStruct goal_predicates = boost::get<PDDLReverseStruct>(*(pddl_reverse_struct_opt->children.begin()+1));
        
        try {
            Goal goal = Goal( parse_predicate_group<Symbol>(goal_predicates), goal_reward );
            goals.push_back(goal);
        }
        catch (PDDLException& e) {
            Goal goal = Goal( parse_predicate_group<TypedSymbol>(goal_predicates), goal_reward );
            goals.push_back(goal);
        }
    }
    
//     LOG(INFO) << "Goals are " << goals;
    return goals;
}

PDDLTypeVector PDDLParser::read_pddl_types(const std::string& str) const
{
    PDDLReverseStructOpt pddl_reverse_struct_opt = find_domain_field(str, PDDL_KEY_TYPES);
    if (pddl_reverse_struct_opt) {
        // remove pddl keyword
        pddl_reverse_struct_opt->children.erase(pddl_reverse_struct_opt->children.begin());
        return  parse_string_list(*pddl_reverse_struct_opt);
    }
    return std::vector<std::string>();
}



/**
 * \class PDDLFileReader
 */

PDDLFileReader::~PDDLFileReader()
{
    if (pddl_file.is_open()) {
        pddl_file.close();
    }
}

void PDDLFileReader::parse_domain_file(RuleSetPtr& ruleset_ptr) const
{
    std::string pddl_domain_file_content = utils::get_file_contents(file_path);
    
    ruleset_ptr = pddl_parser.read_symbolic_ruleset(pddl_domain_file_content);
}

void PDDLFileReader::parse_problem_file(State& state, GoalGroup& goals) const
{
    std::string pddl_domain_file_content = utils::get_file_contents(file_path);
    
    state = pddl_parser.read_state(pddl_domain_file_content);
    goals = pddl_parser.read_global_goals(pddl_domain_file_content);
}

void PDDLFileReader::parse_domain_file(PDDLTypeVector& types) const
{
    std::string pddl_domain_file_content = utils::get_file_contents(file_path);
    
    types = pddl_parser.read_pddl_types(pddl_domain_file_content);
}

void PDDLFileReader::parse_transitions_file(TransitionGroup& transitions)
{
    open_file();
    bool file_finished = false;
    while (!file_finished) {
        boost::optional<std::string> line_prev = read_line();
        boost::optional<std::string> line_action = read_line();
        boost::optional<std::string> line_next = read_line();
        boost::optional<std::string> line_reward = read_line();
        
        if (line_reward) { // checking last one is enough
            PDDLParser pddl_parser;
            PredicateGroup<Symbol> transition_pre = pddl_parser.read_predicate_group(*line_prev);
            TypedSymbol transition_action = pddl_parser.read_typed_symbol(*line_action);
            PredicateGroup<Symbol> transition_next = pddl_parser.read_predicate_group(*line_next);
            float transition_reward = boost::lexical_cast<float>(*line_reward);
            
            Transition transition(transition_pre,
                                transition_action,
                                transition_next,
                                transition_reward);
            transitions.push_back(transition);
        }
        else {
            file_finished = true;
        }
    }
}

void PDDLFileReader::open_file()
{
    pddl_file.open(file_path.c_str());
}

boost::optional<std::string> PDDLFileReader::read_line()
{
    boost::optional<std::string> result;
    
    if (pddl_file.is_open()) {
        std::string line;

        if (pddl_file.good()) {
            getline(pddl_file, line);
        }
        result = line;
    }
    
    return result;
}
