#include <symbolic/predicates.h>

using namespace std;


bool operator==(const Outcome &o1, const Outcome &o2)
{
    return ((*static_cast<const PredicateGroup<Symbol>*>(&o1) == o2) && (o1.probability == o2.probability));
}

bool operator==(const OutcomeList &o1, const OutcomeList &o2)
{
    bool result = true;
    OutcomeList o2_copy(o2);
    for (OutcomeList::const_iterator out_it = o1.begin() ; out_it != o1.end(); out_it++) {
        bool present = false;
        for (OutcomeList::iterator out_it2 = o2_copy.begin() ; out_it2 != o2_copy.end();) {
            if (*out_it == *out_it2) {
                present = true;
                out_it2 = o2_copy.erase(out_it2);
            } else {
                ++out_it2;
            }
        }
        if (!present) {
            result = false;
            break;
        }
    }
    if (o2_copy.size() != 0) { // all outcomes should have been removed
        result = false;
    }
    return result;
}

std::ostream& operator<<(std::ostream &out, const Outcome& o)
{
    out << o.probability << " ";
    if (o.noisy) {
        out << "<noise>";
    } else {
        out << *static_cast<const PredicateGroup<Symbol>*>(&o);
    }
    return out;
}

bool operator<(const Outcome& l, const Outcome& r)
{
    if (*static_cast<const PredicateGroup<Symbol>*>(&l) < *static_cast<const PredicateGroup<Symbol>*>(&r))
        return true;
    else if (*static_cast<const PredicateGroup<Symbol>*>(&r) < *static_cast<const PredicateGroup<Symbol>*>(&l))
        return false;
    else if (l.probability < r.probability)
        return true;
    else if (r.probability < l.probability)
        return false;
    else
        return false;
}

bool operator<(const OutcomeGroup& l, const OutcomeGroup& r)
{
    if (l.size() < r.size())
        return true;
    else if (r.size() < l.size())
        return false;

    OutcomeGroup l_copy(l);
    OutcomeGroup r_copy(r);
    std::sort(l_copy.begin(), l_copy.end());
    std::sort(r_copy.begin(), r_copy.end());

    OutcomeGroup::iterator l_it = l_copy.begin();
    for (OutcomeGroup::iterator r_it = r_copy.begin(); r_it != r_copy.end(); ++r_it) {
        if (*l_it < *r_it)
            return true;
        else if (*r_it < *l_it)
            return false;
        ++l_it;
    }
    return false;
}

void Outcome::write_PPDDL(std::ostream &out) const
{
    out << " " << probability << " ";
    (static_cast<const PredicateGroup<Symbol>*>(this))->write_PPDDL(out);
}

void OutcomeGroup::write_PPDDL(std::ostream &out) const
{
    out << "(probabilistic ";
    BOOST_FOREACH(Outcome outcome, *this) {
        if (outcome.is_useful()) {
            outcome.write_PPDDL(out);
        }
    }
    out << ")";
}




/**
 * Outcome
 */
double Outcome::probability_obtain_predicate(const Predicate<Symbol> predicate) const
{
    if (has_predicate(predicate))
        return probability;
    else
        return 0.0;
}


// Other


Predicate<Symbol> get_predicate_from_string(std::string item)
{
    std::string aux;
    bool positive;

    item = utils::trim(item);
//     std::cout << "Generating predicate for: " << item << std::endl;

    aux = item.substr(0, 1);
    if (aux.compare("-") == 0) {
        item = item.substr(1);
        positive = false;
    } else
        positive = true;

    return Predicate<Symbol>(Symbol(item), positive);
}



Outcome::Outcome(std::string line):
    is_dangerous(false)
{
    std::string item;
    noisy = false;

    // remove leading spaces
    line = line.substr(line.find_first_not_of(" \t"));
    std::stringstream ss(line);

    if (std::getline(ss, item, ' ')) {
        //std::cout << "Probability: " << item << std::endl;
    }
    probability = atof(item.c_str());

    if (ss.eof())
        return;

    // get rest of the line
    std::getline(ss, item);
    if (item.find("<noise>") != std::string::npos) {
        noisy = true;
    } else if (item.find("<no-change>") != std::string::npos) {

    } else {
        replace_predicates(PredicateGroup(item));
    }
}

bool Outcome::changes_predicates() const
{
    if ((size() > 0) && (!noisy)) {
        return true;
    } else {
        return false;
    }
}

bool Outcome::is_useful() const
{
    if (changes_predicates() && (probability > 0.0)) {
        return true;
    } else {
        return false;
    }
}

bool Outcome::apply(PredicateGroup& input_predicates) const
{
    if (changes_predicates()) {
        input_predicates.update_predicates(*this);
        return true;
    } else
        return false;
}

bool OutcomeGroup::changes_predicates() const
{
    bool result = false;
    for (OutcomeGroup::const_iterator outcome_it = this->begin(); outcome_it != this->end(); outcome_it++) {
        if (outcome_it->changes_predicates()) {
            result = true;
        }
    }
    return result;
}

/** \brief aux function to sort in OutcomeGroup::reorder_by_probability */
bool has_greater_outcome (Outcome i,Outcome j) { return (i.probability>j.probability); }

void OutcomeGroup::reorder_by_probability()
{   
    std::sort(begin(), end(), has_greater_outcome);
}

void OutcomeGroup::transform_symbolic_NID_to_symbolic_PDDL()
{
    BOOST_FOREACH(Outcome& outcome, *this) {
        outcome.transform_symbolic_NID_to_symbolic_PDDL();
    }
}

void OutcomeGroup::add_new_outcomes(OutcomeGroup& new_outcomes)
{
    insert( end(), new_outcomes.begin(), new_outcomes.end() );
}

void OutcomeGroup::add_empty_outcome_to_fill_probability()
{
    float total_probability = 0.0;
    BOOST_FOREACH(const Outcome& outcome, *this) {
        total_probability += outcome.probability;
    }
    if (total_probability < 1.0) {
        push_back(Outcome(1.0 - total_probability));
    }
}


// NON-CLASS FUNCTIONS

PredicateGroup<Symbol> read_predicates_from_file(const std::string file_path)
{
    LOG(WARNING) << "Deprecated";
    std::string predicates_string;

    std::ifstream predicates_file(file_path.c_str());
    if (predicates_file.is_open()) {
        if (predicates_file.good()) {
            std::getline(predicates_file, predicates_string);
        }
        predicates_file.close();
    } else LOG(ERROR) << "Unable to open file: " << file_path;

    return PredicateGroup<Symbol>(predicates_string);
}





