/*
 * Clases representing predicates and goals.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_PREDICATES_H
#define IRI_RULE_LEARNER_PREDICATES_H 1


#include <sys/types.h>
#include <vector>
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>

#include <boost/lexical_cast.hpp>

#include <symbolic/symbols.h>

/*!
\class Predicate
\brief A Predicate is a Symbol with a sign.
*/
template<typename SymbolT>
class Predicate : public SymbolT
{
public:
    typedef boost::shared_ptr<Predicate<SymbolT> > Ptr;
    typedef boost::shared_ptr<const Predicate<SymbolT> > ConstPtr;
    
    bool positive;

    Predicate(const SymbolT& symbol_t, const bool positive_):
        SymbolT(symbol_t),
        positive(positive_)
    { }

    /**
     * \brief obtains predicate value as a string
     * 
     * WARNING: At this moment only bool values are supported
     */
    std::string get_value() const {
        if (positive)
            return "true";
        else
            return "false";
    }

    void negate() {
        positive = !positive;
    }

    Predicate get_negated() const {
        return Predicate(*this, !positive);
    }
    
    /**
     * \brief gets the predicates with all the possible parameter permutations given a list possible parameters
     * 
     * \param symbolic_params params from which permutations of parameters are created. For every permutation a new predicate is created. Default parameters are added if there are not enough parameters to fill the predicate parameters.
     * \return predicates having all the permutations as parameters
     */
    std::vector<Predicate> get_symbolic_predicates_permutations(const std::vector<std::string> symbolic_params) const;

    void write_PPDDL(std::ostream &out) const;

    /**
     * \brief compares two predicates as if they were symbolic, so parameter names are not compared
     */
    bool symbolic_compare(const Predicate& p) const;
    
//     friend bool operator==(const Predicate<SymbolT> &p1, const Predicate<SymbolT> &p2);
//     friend bool operator<(const Predicate<SymbolT>& l, const Predicate<SymbolT>& r);
//     friend std::ostream& operator<<(std::ostream &out, const Predicate<SymbolT>& p);
};



/*!
\class PredicateGroup
\brief A vector of Predicate.
*/
template<typename SymbolT>
class PredicateGroup 
{
private:
    std::vector< Predicate<SymbolT> > internal_vector;
    
public:
    /** Begin of methods exposing std::vector stuff */
    typedef Predicate<SymbolT> value_type;
    typedef size_t size_type;
    typedef size_t index_type;
    // iterators
    typedef typename std::vector< value_type >::iterator iterator;
    typedef typename std::vector< value_type >::const_iterator const_iterator;
    iterator        begin()         { return internal_vector.begin(); };
    const_iterator  begin() const   { return internal_vector.begin(); };
    iterator        end()           { return internal_vector.end(); };
    const_iterator  end() const     { return internal_vector.end(); };
    
    // change elements
    void        push_back (const value_type& val)   { internal_vector.push_back(val); };
    // clear
    void clear() { internal_vector.clear(); };
    // erase
    iterator    erase (iterator position)               { return internal_vector.erase(position); };
    iterator    erase (iterator first, iterator last)   { return internal_vector.erase(first, last); };
    // insert
    iterator    insert (iterator position, const value_type& val)                   { return internal_vector.insert(position, val); };
    void        insert (iterator position, size_type n, const value_type& val)      { internal_vector.insert(position, n, val); };
    template <class InputIterator>
    void        insert (iterator position, InputIterator first, InputIterator last) { internal_vector.insert(position, first, last); };
    
    // access elements
    value_type& front()             { return internal_vector.front(); };
    const value_type& front() const { return internal_vector.front(); };
    value_type& operator[](index_type idx)              { return internal_vector[idx]; };
    const value_type& operator[](index_type idx) const  { return internal_vector[idx]; };
        
    // info
    bool empty() const { return internal_vector.empty(); };
    size_type size() const { return internal_vector.size(); }
    std::vector< Predicate<SymbolT> > get_internal_vector() { return internal_vector; }
    /** End of methods exposing std::vector stuff */
    
    
    PredicateGroup() { }
    PredicateGroup(const PredicateGroup<SymbolT>& pred_group) :
        internal_vector(pred_group.internal_vector)
    { }
    PredicateGroup(const std::vector< Predicate<SymbolT> >& predicates) :
        internal_vector(predicates)
    { }
    PredicateGroup(std::string predicates_);

    /**
     * \brief checks if current_predicates satisfy all predicates in PredicateGroup
     * 
     * \return True if PredicateGroup is satisfied by current_predicates
     */
    bool is_satisfied_by(const PredicateGroup current_predicates) const;

    /**
     * \brief checks if PredicateGroup satisfies the Predicate predicate
     * 
     * \return True if predicate is satisfied by PredicateGroup
     */
    bool satisfies(const Predicate<SymbolT> predicate) const;

    /**
     * \brief adds predicate to the state and removes duplicates
     */
    inline void add_predicates(const PredicateGroup<SymbolT> new_predicates) {
        this->insert(this->end(), new_predicates.begin(), new_predicates.end());
        remove_duplicated();
    }
    
    /**
     * \brief Removes current predicates and adds the new ones.
     */
    inline void replace_predicates(const PredicateGroup new_predicates) {
        this->clear();
        add_predicates(new_predicates);
    }

    /**
     * \brief Count number of predicates satisfied by current_predicates
     */
    int count_satisfied_by(PredicateGroup current_predicates) const;

    PredicateGroup get_not_satisfied_by(const PredicateGroup predicates) const {
        PredicateGroup<SymbolT> predicates_not_satisfied;
        BOOST_FOREACH(const Predicate<SymbolT>& predicate_it, *this) {
            if ( !predicates.satisfies(predicate_it) ) {
                predicates_not_satisfied.push_back(predicate_it);
            }
        }
        return predicates_not_satisfied;
    }
    
    bool has_predicate(const Predicate<SymbolT>& predicate) const;

    void change_variable(const std::string& orig_var, const std::string& new_var);
    
    /**
     * \brief Symbolic params are grounded with grounded params
     */
    void ground_params(const std::vector< std::string >& symbolic_params_, const std::vector< std::string >& grounded_params_);
    
    /**
     * \brief Adds new_predicates and removes their negated counterparts if exists
     */
    void update_predicates(const PredicateGroup& new_predicates);
    
    /**
     * \brief Adds new_predicate and removes its negated counterpart if exists
     */
    void update_predicate(const Predicate<SymbolT>& new_predicate);

    /*!
     * \brief Makes the predicates symbolic
     *
     * Changes the params to be the default symbolic params: X, Y, Z, V, W, Q.
     */
    void make_symbolic() { BOOST_FOREACH(Predicate<SymbolT>& predicate, *this){predicate.make_symbolic();} }
    
    std::vector<std::string> get_variables() const;

    // Comparisons and differences
    /**
     * \brief Add differences with the compared group that are positive
     */
    void add_positive_differences(PredicateGroup compared_group);

    /**
     * \brief Get differences with compared PredicateGroup
     *
     *  Adds predicates in compared_group that are not present in current predicates.
     * Predicates not present in compared_group will be added negated.
     */
    PredicateGroup get_differences(const PredicateGroup compared_group) const;

    /**
     * \brief gets the predicates with all the possible parameter permutations given a list possible parameters
     * 
     * \param param_names params from which permutations of parameters are created. For every permutation a new predicate is created. Default parameters are added if there are not enough parameters to fill the predicate parameters.
     * \return predicates having all the permutations as parameters
     */
    PredicateGroup get_symbolic_predicates_permutations(const std::vector< std::string >& param_names) const;
    
    /** \brief Transforms X to ?X, Y to ?Y, etc... */
    void transform_symbolic_NID_to_symbolic_PDDL();

    void remove(const PredicateGroup<SymbolT> predicates_to_remove);
    
    void remove(const Predicate<SymbolT> predicate);
    
    
    /** \brief Remove repeated predicates */
    void remove_duplicated();
    /** \brief Remove predicates whose params are not in list_of_params  */
    void remove_predicates_with_params_not_in(std::vector<std::string> list_of_params);

    /** \brief Gets predicates negated */
    PredicateGroup get_negatet_predicates() const;
    /** \brief Removes negated predicates */
    void remove_negated_predicates();

    /** \brief Does a PPDDL grounding for IPPC */
    void PPDDL_grounding();
    /** \brief Writes in standard PPDDL format */
    void write_PPDDL(std::ostream &out) const;
    /** \brief Writes in PPDDL format without leading and */
    void write_simplified_PPDDL(std::ostream &out) const;
    /** \brief Writes (or (not p1) (not p2) ... ) */
    void write_negated_PPDDL(std::ostream &out) const;

    void write_to_file(const std::string file_path) const;

    /**
     * Operators
     */
//     friend std::ostream& operator<<(std::ostream &out, const PredicateGroup<SymbolT>& p);
//     friend bool operator==(const PredicateGroup<SymbolT> &p1, const PredicateGroup<SymbolT> &p2);
//     friend bool operator<(const PredicateGroup<SymbolT>& l, const PredicateGroup<SymbolT>& r);
    
private:
    bool has_symbolic_predicate(const Predicate<SymbolT>& predicate) const;
    bool has_grounded_predicate(const Predicate<SymbolT>& predicate) const;
};



/*!
\class Outcome
\brief A PredicateGroup with an associated probability

The predicates define the changes made to the state when the outcome is applied. The probability defines the chances that this outcome is selected in a rule.
*/
class Outcome : public PredicateGroup<Symbol>
{
public:
    float probability;
    bool noisy;
    bool is_dangerous;

    /**
     * Constructors
     */
    Outcome(std::string line);

    Outcome(float probability_, bool noisy_ = false) :
        PredicateGroup(),
        probability(probability_),
        noisy(noisy_),
        is_dangerous(false)
    { }

    Outcome(PredicateGroup predicates_, float probability_) :
        PredicateGroup(predicates_),
        probability(probability_),
        noisy(false),
        is_dangerous(false)
    { }

    /**
     * Helpers
     */
    double probability_obtain_predicate(const Predicate<Symbol> predicate) const;
    /** \brief true if changes any predicate */
    bool changes_predicates() const;
    /** \brief true if changes any predicate and has a probability > 0.0 */
    bool is_useful() const;
    /** \brief Changes predicates by updating them with its PredicateGroup */
    bool apply(PredicateGroup& predicates) const;

    void write_PPDDL(std::ostream &out) const;

    /**
     * Operators
     */
    friend std::ostream& operator<<(std::ostream &out, const Outcome& o);
    friend bool operator==(const Outcome &o1, const Outcome &o2);
    friend bool operator<(const Outcome& l, const Outcome& r);
};

typedef boost::optional<Outcome> OutcomeOpt;
typedef std::vector<Outcome> OutcomeList;
bool operator==(const OutcomeList &o1, const OutcomeList &o2);

/*!
\class OutcomeGroup
\brief A vector of outcomes
*/
class OutcomeGroup : public OutcomeList
{
public:
    /** \brief true if changes any predicate */
    bool changes_predicates() const;
    
    /** \brief Reorders outcomes by decreasing probability */
    void reorder_by_probability();
    
    /** \brief Transforms X to ?X, Y to ?Y, etc... */
    void transform_symbolic_NID_to_symbolic_PDDL();
    
    /** \brief Add new outcomes */
    void add_new_outcomes(OutcomeGroup& new_outcomes);
    
    /** \brief Add empty outcome to fill probabilities up to 1.0 */
    void add_empty_outcome_to_fill_probability();
    
    void write_PPDDL(std::ostream &out) const;

    friend bool operator<(const OutcomeGroup& l, const OutcomeGroup& r);  // WARNING does not compare probabilities
};


// get data from string
Predicate<Symbol> get_predicate_from_string(std::string item);
PredicateGroup<Symbol> read_predicates_from_file(const std::string file_path);

#include <symbolic/predicates_template_imp.h>

#endif

