/*
 * Template implementations for predicates.h.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


#ifndef IRI_RULE_LEARNER_PREDICATES_TEMPLATE_IMPL_H
#define IRI_RULE_LEARNER_PREDICATES_TEMPLATE_IMPL_H 1

#include <symbolic/predicates.h>

// Operators
template<typename SymbolT>
bool operator==(const Predicate<SymbolT> &p1, const Predicate<SymbolT> &p2)
{
    return ((*static_cast<const Symbol*>(&p1) == p2) && (p1.positive == p2.positive));
}

template<typename SymbolT>
bool Predicate<SymbolT>::symbolic_compare(const Predicate<SymbolT>& p) const
{
    return ((static_cast<const SymbolT*>(&*this)->symbolic_compare(p)) && (positive == p.positive));
}

template<typename SymbolT>
bool operator==(const PredicateGroup<SymbolT> &p1, const PredicateGroup<SymbolT> &p2)
{
    bool result = true;
    PredicateGroup<SymbolT> p2_copy(p2);
    for (typename PredicateGroup<SymbolT>::const_iterator it_pred = p1.begin() ; it_pred != p1.end(); it_pred++) {
        bool present = false;
        for (typename PredicateGroup<SymbolT>::iterator it_pred2 = p2_copy.begin() ; it_pred2 != p2_copy.end();) {
            if (*it_pred == *it_pred2) {
                present = true;
                it_pred2 = p2_copy.erase(it_pred2);
            } else {
                ++it_pred2;
            }
        }
        if (!present) {
            result = false;
            break;
        }
    }
    if (p2_copy.size() != 0) { // all predicates should have been removed
        result = false;
    }
    return result;
}

template<typename SymbolT>
std::ostream& operator<<(std::ostream &out, const Predicate<SymbolT>& p)
{
    if (not p.positive)
        out << "-";
    out << *static_cast<const SymbolT*>(&p);
    return out;
}

template<typename SymbolT>
std::ostream& operator<<(std::ostream &out, const PredicateGroup<SymbolT>& p)
{
    for (size_t i = 0; i < p.size(); i++) {
        out << p[i];
        if ((i + 1) < p.size())
            out << " ";
    }
    return out;
}

template<typename SymbolT>
bool operator<(const Predicate<SymbolT>& l, const Predicate<SymbolT>& r)
{
    if (*static_cast<const SymbolT*>(&l) < *static_cast<const SymbolT*>(&r))
        return true;
    else if (*static_cast<const SymbolT*>(&r) < *static_cast<const SymbolT*>(&l))
        return false;
    else if (l.positive < r.positive)
        return true;
    else
        return false;
}

template<typename SymbolT>
bool operator<(const PredicateGroup<SymbolT>& l, const PredicateGroup<SymbolT>& r)
{
    if (l.size() < r.size())
        return true;
    else if (r.size() < l.size())
        return false;

    PredicateGroup<SymbolT> l_copy(l);
    PredicateGroup<SymbolT> r_copy(r);
    std::sort(l_copy.begin(), l_copy.end());
    std::sort(r_copy.begin(), r_copy.end());

    typename PredicateGroup<SymbolT>::const_iterator l_it = l_copy.begin();
    for (typename PredicateGroup<SymbolT>::const_iterator r_it = r_copy.begin(); r_it != r_copy.end(); ++r_it) {
        if (*l_it < *r_it)
            return true;
        else if (*r_it < *l_it)
            return false;
        ++l_it;
    }
    return false;
}

template<typename SymbolT>
void Predicate<SymbolT>::write_PPDDL(std::ostream &out) const
{
    if (!positive) {
        out << "(not ";
    }
    (static_cast<const SymbolT*>(this))->write_PPDDL(out);
    if (!positive) {
        out << ")";
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::write_PPDDL(std::ostream &out) const
{
    // Maximun of 15 predicates per line (to not exceed character limit in editors)
    uint number_predicates = 0;
    out << "(and ";
    for (typename PredicateGroup<SymbolT>::const_iterator it_pred = this->begin() ; it_pred != this->end(); ++it_pred) {
        it_pred->write_PPDDL(out);
        out << " ";
        if (number_predicates++ > 20) { // max predicates = 20 (performance issues)
            number_predicates = 0;
            out << std::endl;
        }
    }
    out << ")";
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::write_negated_PPDDL(std::ostream &out) const
{
    // Maximun of 15 predicates per line (to not exceed character limit in editors)
    uint number_predicates = 0;
    out << "(or ";
    for (typename PredicateGroup<SymbolT>::const_iterator it_pred = this->begin() ; it_pred != this->end(); ++it_pred) {
        it_pred->get_negated().write_PPDDL(out);
        out << " ";
        if (number_predicates++ > 20) { // max predicates = 20 (performance issues)
            number_predicates = 0;
            out << std::endl;
        }
    }
    out << ")";
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::write_simplified_PPDDL(std::ostream &out) const 
{
    BOOST_FOREACH(const Predicate<SymbolT>& predicate, *this) {
        predicate.write_PPDDL(out);
        out << " ";
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::PPDDL_grounding()
{
    for (typename PredicateGroup<SymbolT>::iterator it_pred = this->begin() ; it_pred != this->end(); ++it_pred) {
        it_pred->PPDDL_grounding();
    }
}



/// Predicate
template<typename SymbolT>
std::vector<Predicate<SymbolT> > Predicate<SymbolT>::get_symbolic_predicates_permutations(const std::vector<std::string> symbolic_params) const {
    PredicateGroup<Symbol> grounded_predicates;
    std::vector<std::string> full_symbolic_params(symbolic_params);
    while (full_symbolic_params.size() < this->param_names.size()) { // add default parameters if there are not enough of them
        full_symbolic_params.push_back(DEFAULT_PARAM_NAMES[this->param_names.size()]);
    }
    std::sort(full_symbolic_params.begin(), full_symbolic_params.end());

    do {
        Predicate<SymbolT> predicate(*this);
        for (size_t j = 0; j < this->param_names.size(); j++) {
            predicate.param_names[j] = full_symbolic_params[j];
        }
        grounded_predicates.push_back(predicate);
    } while (std::next_permutation(full_symbolic_params.begin(), full_symbolic_params.end()));

    grounded_predicates.remove_duplicated();
    return grounded_predicates.get_internal_vector();
}

/**
 * ***************
 * PredicateGroup
 */

// PredicateGroup

template<typename SymbolT>
PredicateGroup<SymbolT>::PredicateGroup(std::string predicates_)
{
    // DEPRECATED
    std::string item, aux, line;
    bool positive;

    // remove leading and trailing spaces
    line = utils::trim(predicates_);

    std::stringstream ss(line);

    // case of empty string
    if ((line == "") || (line == "--")) {
        return;
    }

    while (std::getline(ss, item, ')')) {
        item = utils::trim(item);
//         std::cout << "Generating predicate for: " << item << std::endl;

        if (item.size() > 1) { // if something remains
            // add ) to the end to complete it
            item = item + ')';

            // if begins with , remove it
            aux = item.substr(0, 1);
            if (aux.compare(",") == 0)
                item = item.substr(1);

            item = utils::trim(item);
            aux = item.substr(0, 1);
            if (aux.compare("-") == 0) {
                item = item.substr(1);
                positive = false;
            } else
                positive = true;

            this->push_back(Predicate<SymbolT>(Symbol(item), positive));
        }
    }
}

template<typename SymbolT>
bool PredicateGroup<SymbolT>::satisfies(const Predicate<SymbolT> predicate_to_satisfy) const
{
    bool result = false;
    bool negated_pred_present = false;
    
    BOOST_FOREACH(const Predicate<SymbolT>& pred_it, *this) {
        if ( predicate_to_satisfy == pred_it ) {
            result = true;
        }
        if ( predicate_to_satisfy == pred_it.get_negated() ) {
            negated_pred_present = true;
        }
    }
    // negated predicates don't have to be present (closed-world assumption)
    if ((!predicate_to_satisfy.positive) && (!negated_pred_present)) {
        result = true;
    }
    return result;
}

template<typename SymbolT>
bool PredicateGroup<SymbolT>::is_satisfied_by(const PredicateGroup<SymbolT> current_predicates) const
{
    bool satisfied = true;
    BOOST_FOREACH(const Predicate<SymbolT>& predicate_it, *this) {
        if ( !current_predicates.satisfies(predicate_it) ){
            satisfied = false;
            break; // ugly but quicker
        }
    }
//     std::cout << "Checking predicates\nPreconditions:" << *this << "\nState:" << PredicateGroup(current_predicates) << "\nResult: " << satisfied << std::endl;
    return satisfied;
}

template<typename SymbolT>
int PredicateGroup<SymbolT>::count_satisfied_by(const PredicateGroup<SymbolT> current_predicates) const
{
    int n_satisfied = 0;
    BOOST_FOREACH(const Predicate<SymbolT>& predicate_it, *this) {
        if ( current_predicates.satisfies(predicate_it) ){
            n_satisfied++;
        }
    }
//         std::cout << "Counting predicates\nPreconditions:" << *this << "\nState:" << PredicateGroup(current_predicates) << "\nResult: " << n_satisfied << std::endl;
    return n_satisfied;
}

template<typename SymbolT>
bool PredicateGroup<SymbolT>::has_predicate(const Predicate<SymbolT>& predicate) const
{
    BOOST_FOREACH(const Predicate<SymbolT>& predicate_it, *this) {
        if (predicate.is_symbolic || predicate_it.is_symbolic) {
            if ( has_symbolic_predicate(predicate) ) {
                return true;
            }
        }
        else {
            if ( has_grounded_predicate(predicate) ) {
                return true;
            }
        }
    }
    return false;
}

template<typename SymbolT>
std::vector<std::string> PredicateGroup<SymbolT>::get_variables() const
{
    std::vector<std::string> result;
    std::set<std::string> variables;
    BOOST_FOREACH(const Predicate<SymbolT>& predicate, *this) {
        BOOST_FOREACH(const std::string& param, predicate.param_names) {
            variables.insert(param);
        }
    }
    result.assign(variables.begin(), variables.end());
    return result;
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::add_positive_differences(PredicateGroup compared_group){
    PredicateGroup<SymbolT> differences = compared_group.get_differences(*this);
    BOOST_FOREACH(const Predicate<SymbolT>& predicate, differences) {
        if (predicate.positive) {
            this->push_back(predicate);
        }
    }
    remove_duplicated();
}

template<typename SymbolT>
bool PredicateGroup<SymbolT>::has_grounded_predicate(const Predicate<SymbolT>& predicate) const
{
    BOOST_FOREACH(const Predicate<SymbolT>& predicate_it, *this) {
        if (predicate_it == predicate) {
            return true;
        }
    }
    return false;
}

template<typename SymbolT>
bool PredicateGroup<SymbolT>::has_symbolic_predicate(const Predicate<SymbolT>& predicate) const
{
    BOOST_FOREACH(const Predicate<SymbolT>& predicate_it, *this) {
        if (predicate_it.symbolic_compare(predicate)) {
            return true;
        }
    }
    return false;
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::change_variable(const std::string& orig_var, const std::string& new_var)
{
    for (typename PredicateGroup<SymbolT>::iterator it_pred = this->begin() ; it_pred != this->end(); it_pred++) {
        for (std::vector<std::string>::iterator it_param = it_pred->param_names.begin(); it_param != it_pred->param_names.end(); it_param++) {
            if (it_param->compare(orig_var) == 0) {
                *it_param = new_var;
            }
        }
        it_pred->autodetect_symbolic();
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::update_predicate(const Predicate<SymbolT>& new_predicate)
{
    bool is_present = false;
    BOOST_FOREACH(Predicate<SymbolT>& predicate, *this) {
        if (*static_cast<const SymbolT*>(&new_predicate) == predicate) { // if equal name and param names
            predicate.positive = new_predicate.positive;
            is_present = true;
        }
    }
    if (!is_present) {
        this->push_back(new_predicate);
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::update_predicates(const PredicateGroup<SymbolT>& new_predicates)
{
    BOOST_FOREACH(const Predicate<SymbolT> & predicate, new_predicates) {
        update_predicate(predicate);
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::remove_duplicated()
{
    for (typename PredicateGroup<SymbolT>::iterator it_pred = this->begin() ; it_pred != this->end(); it_pred++) {
        for (typename PredicateGroup<SymbolT>::iterator it_pred2 = (it_pred + 1) ; it_pred2 != this->end();) {
            if (*it_pred == *it_pred2) {
                it_pred2 = this->erase(it_pred2);
            } else {
                ++it_pred2;
            }
        }
    }
}

template<typename SymbolT>
PredicateGroup<SymbolT> PredicateGroup<SymbolT>::get_differences(const PredicateGroup compared_group) const {
    PredicateGroup<SymbolT> compared_preds(compared_group);
    PredicateGroup<SymbolT> original_preds(*this);
    PredicateGroup<SymbolT> differences;
    bool is_present;
    original_preds.remove_duplicated();
    for (typename PredicateGroup<SymbolT>::const_iterator it_pred = original_preds.begin() ; it_pred != original_preds.end(); it_pred++) {
        is_present = false;
        for (typename PredicateGroup<SymbolT>::iterator it_pred2 = compared_preds.begin() ; it_pred2 != compared_preds.end();) {
            if (*it_pred == *it_pred2) {
                is_present = true;
                it_pred2 = compared_preds.erase(it_pred2);
            } else {
                ++it_pred2;
            }
        }
        if (!is_present) {
            differences.push_back(it_pred->get_negated());
        }
    }
    differences.insert(differences.end(), compared_preds.begin(), compared_preds.end());

    differences.remove_duplicated();
    return differences;
}

template<typename SymbolT>
PredicateGroup<SymbolT> PredicateGroup<SymbolT>::get_symbolic_predicates_permutations(const std::vector< std::string >& param_names) const {
    PredicateGroup<SymbolT> result;
    BOOST_FOREACH(const Predicate<SymbolT>& predicate, *this) {
        result.add_predicates(predicate.get_symbolic_predicates_permutations(param_names));
    }
    return result;
}

template<typename SymbolT>
PredicateGroup<SymbolT> PredicateGroup<SymbolT>::get_negatet_predicates() const
{
    PredicateGroup<SymbolT> result(*this);
    BOOST_FOREACH(Predicate<SymbolT>& predicate, result) {
        predicate.negate();
    }
    return result;
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::remove(const PredicateGroup<SymbolT> predicates_to_remove) 
{
    BOOST_FOREACH(const Predicate<SymbolT>& predicate_to_remove, predicates_to_remove) {
        remove(predicate_to_remove);
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::remove(const Predicate<SymbolT> predicate) 
{
    for (typename PredicateGroup<SymbolT>::iterator it_pred = this->begin() ; it_pred != this->end();) {
        if (predicate.symbolic_compare(*it_pred)) {
            it_pred = this->erase(it_pred);
        } else {
            ++it_pred;
        }
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::remove_negated_predicates() 
{
    for (typename PredicateGroup<SymbolT>::iterator it_pred = this->begin() ; it_pred != this->end();) {
        if (!it_pred->positive) {
            it_pred = this->erase(it_pred);
        } else {
            ++it_pred;
        }
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::remove_predicates_with_params_not_in(std::vector<std::string> list_of_params)
{
    for (typename PredicateGroup<SymbolT>::iterator it_pred = this->begin() ; it_pred != this->end();) {
        if (it_pred->params_belong_to(list_of_params)) {
            ++it_pred;
        } else {
            it_pred = erase(it_pred);
        }
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::transform_symbolic_NID_to_symbolic_PDDL()
{
    BOOST_FOREACH(Predicate<SymbolT>& predicate, *this) {
        predicate.transform_nid_params_to_pddl_params();
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::ground_params(const std::vector< std::string >& symbolic_params_, const std::vector< std::string >& grounded_params_)
{
    assert(symbolic_params_.size() <= grounded_params_.size());
    std::vector<std::string>::const_iterator grounded_param_it;
    grounded_param_it = grounded_params_.begin();
    for (std::vector<std::string>::const_iterator param_it = symbolic_params_.begin(); param_it != symbolic_params_.end(); ++param_it) {
        change_variable(*param_it, *grounded_param_it);
        ++grounded_param_it;
    }
}

template<typename SymbolT>
void PredicateGroup<SymbolT>::write_to_file(const std::string file_path) const
{
    int DEBUG = 0;
    std::ofstream predicates_file(file_path.c_str());
    std::cout << "PredicateGroup::" << "write_to_file: " << "Writting predicates to file " << file_path << std::endl;
    if (DEBUG)
        std::cout << "PredicateGroup::" << "write_to_file: " << "Predicates are " << *this << std::endl << std::endl;
    if (predicates_file.is_open()) {
        predicates_file << *this;
    } else
        std::cerr << "PredicateGroup::" << "write_to_file: " << "Couldn't write predicates to " << file_path << std::endl;
    predicates_file.close();
}

#endif
