/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2013  dmartinez <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "symbolic/scenario.h"


Scenario::Scenario(const RuleSet& rules_, const State& state_, const GoalGroup& goals_, 
                   const ConfigReader& config_reader_, 
                   const std::vector<TypedSymbol>& action_list_,
                   const PDDLTypeVector& ppddl_types_
                  ):
    config_reader(config_reader_),
    rules(rules_),
    state(state_),
    goals(goals_),
    action_list(action_list_),
    pddl_types(ppddl_types_),
    current_reward(0)
{
    init();
}

Scenario::Scenario(const RuleSet& rules_, const std::string& pddl_problem_file, const ConfigReader& config_reader_):
    config_reader(config_reader_),
    rules(rules_),
    current_reward(0)
{
    read_ppddl_problem_file(pddl_problem_file);
    init();
}

Scenario::Scenario(const std::string& pddl_domain_file, const std::string& pddl_problem_file, const ConfigReader& config_reader_):
    config_reader(config_reader_),
    current_reward(0)
{
    read_ppddl_problem_file(pddl_problem_file);
    read_ppddl_domain_file(pddl_domain_file);
    init();
}

Scenario::Scenario(const Scenario& scenario):
    domain(scenario.domain),
    config_reader(scenario.config_reader),
    rules(scenario.rules),
    state(scenario.state),
    goals(scenario.goals),
    action_list(scenario.action_list),
    pddl_types(scenario.pddl_types),
    current_reward(scenario.current_reward),
    noop_reward(scenario.noop_reward),
    planning_horizon(scenario.planning_horizon)
{
}

void Scenario::init()
{
    if (pddl_types.empty()) {
        BOOST_FOREACH(const PDDLTypeObjectMap::value_type& type_objects, state.pddl_object_manager.get_ppddl_objects_per_type())
        {
            pddl_types.push_back(type_objects.first);
        }
    }
    
    domain.index_goals(goals);
    domain.index_state(state);
    domain.index_rules(rules);
    
    if (config_reader.get_variable<bool>("final_reward_action")) { // add -finished precondition
        this->rules.add_predicate_to_preconditions(Predicate<Symbol>(MAKE_SYMBOL_FINISHED, false));
    }
    
    if (action_list.empty()) {
        action_list = rules.get_actions();
    }
    
    this->noop_reward = config_reader.get_variable<float>("planning_noop_reward");
    this->planning_horizon = config_reader.get_variable<uint>("planning_horizon");
    
    if (config_reader.get_variable<bool>("final_reward_action")) {
        std::vector<TypedSymbol>::iterator finish_action_it = std::find(action_list.begin(), action_list.end(), MAKE_TYPED_SYMBOL_FINISH_ACTION);
        if (finish_action_it == action_list.end()) { // if finish_action wasn't found, load it manually
            // DEPRECATED This is for compatibility with old way of define domains
            LOG(WARNING) << "Deprecated separate final reward action";
            LOG(ERROR) << START_COLOR_RED << "TODO WARNING FIX THIS!!!" << END_COLOR;
//             RuleSet finish_rule(config_reader.get_path("nid_finish_rule_path"));
//             rules.add_rules(finish_rule);
//             domain.index_rules(finish_rule);
            // -finished() predicate used by final reward action
//             rules.add_predicate_to_preconditions(Predicate(MAKE_SYMBOL_FINISHED, false));
        }
    }
}

void Scenario::add_Rules_to_action_list(const RuleSet& rules)
{
    std::vector<TypedSymbol> actions = rules.get_actions();
    BOOST_FOREACH(const TypedSymbol& action, actions) {
        add_action_to_action_list(action);
    }
}


void Scenario::add_action_to_action_list(TypedSymbol action)
{
    action.make_symbolic();
    bool isPresent = (std::find(action_list.begin(), action_list.end(), action) != action_list.end());
    if (!isPresent) {
        action_list.push_back(action);
    }
}

void Scenario::add_action_to_action_list(Symbol action)
{
    add_action_to_action_list(state.pddl_object_manager.get_typed_symbol(action));
}

void Scenario::add_actions_to_action_list(std::vector<TypedSymbol> action_list)
{
    BOOST_FOREACH(const TypedSymbol& action, action_list) {
        add_action_to_action_list(action);
    }
}

// needs refactoring
// It supposes that the first outcome is the successful one
bool Scenario::apply(const GroundedRule& grounded_rule)
{
    bool aux_success;
    float aux_reward;
    return apply(grounded_rule, aux_success, aux_reward);
}


bool Scenario::apply(const GroundedRule& grounded_rule, bool& success, float& reward, const bool force_outcome, int force_outcome_idx)
{
    // first, apply global rewards
    reward = goals.obtain_reward(state);
    
    // second, apply rule reward before effects. 
    // Note that reward is only obtained if goal preconditions are satisfied (goal preconditions have to be properly defined)
    reward += grounded_rule.goals.obtain_reward(state);
    
    // third, apply rule effects
    bool result = false;
    if ((!grounded_rule.preconditions.is_satisfied_by(state)) || (!grounded_rule.outcomes.changes_predicates())) { 
        result = false;
    } else if (force_outcome) {
        if (force_outcome_idx < 0) { // autoselect outcome
            force_outcome_idx = 0;
            while ( (force_outcome_idx < (int)grounded_rule.outcomes.size()) && grounded_rule.outcomes[force_outcome_idx].empty() ) {
                force_outcome_idx++;
            }
        }
        if (force_outcome_idx < (int)grounded_rule.outcomes.size()) { // check if valid outcome was selected
            grounded_rule.outcomes[force_outcome_idx].apply(state);
        }
        success = true;
        result = true;
    } else {
        double rand_01;
        rand_01 = ((double) rand() / (RAND_MAX));

        double accumulated_prob = 0.0;

        OutcomeList::const_iterator outcome_it = grounded_rule.outcomes.begin();
        size_t out_idx = 1;
        accumulated_prob += outcome_it->probability;
        while (accumulated_prob < rand_01) {
            ++ outcome_it;
            ++out_idx;
            accumulated_prob += outcome_it->probability;
        }
        outcome_it->apply(state);
        if (out_idx == 1) // first outcome is the successful one
            success = true;
        else
            success = false;

        result = true;
    }
    
    return result;
}


bool Scenario::apply_action(Symbol action)
{
    bool success;
    float reward;
    return apply_action(action, success, reward);
}

bool Scenario::apply_action(Symbol action, bool& success, float& reward, const bool force_outcome, const int force_outcome_idx)
{
    success = false;
    bool rules_applied = false;
    int number_rules_applied = 0;
    RuleSet grounded_action_rules = rules.get_grounded_action_rules(action, state.pddl_object_manager);
    BOOST_FOREACH(RulePtr rule_ptr, grounded_action_rules) {
        GroundedRulePtr grounded_rule_ptr = boost::make_shared<GroundedRule>(*static_cast<const GroundedRule*>(&(*rule_ptr)));
        if (apply(*grounded_rule_ptr, success, reward, force_outcome, force_outcome_idx)) {
            number_rules_applied++;
            rules_applied = true;
        }
    }

    if (number_rules_applied > 1) {
        std::cerr << "RuleSet: Rules bad defined, more than one rule was applied with action " << action << "." << std::endl;
    }
    
    current_reward += reward;

    return rules_applied;
}

void Scenario::read_ppddl_domain_file(const std::string& ppddl_file_path)
{
    PDDLFileReader pddl_file_reader(ppddl_file_path);
    // types
    pddl_file_reader.parse_domain_file(this->pddl_types);
    //rules
    RuleSetPtr rules_ptr;
    pddl_file_reader.parse_domain_file(rules_ptr);
    this->rules = *rules_ptr;
}

void Scenario::read_ppddl_problem_file(const std::string& ppddl_file_path)
{
    PDDLFileReader pddl_file_reader(ppddl_file_path);
    pddl_file_reader.parse_problem_file(this->state, this->goals);
}

void Scenario::write_PDDL_domain(const std::string& file_path)
{
    std::stringstream ss;
    write_PDDL_domain(ss);
    utils::write_string_to_file(file_path, ss.str());
}

void Scenario::write_PDDL_problem(const std::string& file_path)
{
    std::stringstream ss;
    write_PDDL_problem(ss);
    utils::write_string_to_file(file_path, ss.str());
}

void Scenario::write_PDDL_domain(std::ostream& out)
{
    out << "(define (domain " << config_reader.get_variable<std::string>("domain") << ")" << std::endl;
    out << "  (:requirements :typing :strips :equality :probabilistic-effects :rewards)" << std::endl;
    // types
    out << "  (:types";
    BOOST_FOREACH(PDDLType pddl_type, pddl_types) {
        out << " " << pddl_type;
    }
    out << ")" << std::endl;
    // predicates
    domain.write_PPDDL_symbolic_predicates(out, state.pddl_object_manager);
    // rules
    rules.write_PPDDL(out);
    out << ")" << std::endl;
}

void Scenario::write_PDDL_problem(std::ostream& out)
{
    out << "(define (problem " << config_reader.get_variable<std::string>("problem") << ")" << std::endl;
    out << "  (:domain " << config_reader.get_variable<std::string>("domain") << ")" << std::endl;
    // objects
    out << "  (:objects ";
    BOOST_FOREACH(const PDDLTypeObjectMap::value_type& objects_per_type, state.pddl_object_manager.get_ppddl_objects_per_type()) {
        BOOST_FOREACH(const std::string& object, objects_per_type.second) {
            out << object << " ";
        }
        out << " - ";
        out << objects_per_type.first << " ";
    }
    out << ")" << std::endl;
    // init
    out << "  (:init ";
    state.write_simplified_PPDDL(out);
    out << ")" << std::endl;
    BOOST_FOREACH(const Goal& goal, goals) {
        // goal
        out << "  (:goal ";
        goal.get_grounded_predicates().write_PPDDL(out);
        out << ")" << std::endl;
        // goal reward
        out << "  (:goal-reward " << goal.get_reward() << ")" << std::endl;
    }
    // metric
    out << "  (:metric maximize (reward))" << std::endl;
    out << ")" << std::endl;
}


