#include <symbolic/symbols.h>


// OPERATORS
bool operator== (const Symbol &s1, const Symbol &s2)
{
    return ((s1.name == s2.name) && utils::compare_ordered_vector_strings(s1.param_names, s2.param_names));
}

bool operator== (const TypedSymbol &s1, const TypedSymbol &s2)
{
    return ((s1.name == s2.name) && utils::compare_ordered_vector_strings(s1.param_names, s2.param_names) && utils::compare_ordered_vector_strings(s1.param_types, s2.param_types));
}

bool operator!= (const Symbol &s1, const Symbol &s2)
{
    return !(s1 == s2);
}

bool operator!= (const TypedSymbol &s1, const TypedSymbol &s2)
{
    return !(s1 == s2);
}

bool Symbol::symbolic_compare(const Symbol &s) const
{
    int DEBUG = 0;
    bool equal = (s.name == name);
    if (DEBUG > 2)
        LOG(INFO) << "Symbolic " << *this << ": " << is_symbolic << " - Symbolic " << s << ": " << s.is_symbolic;
    if (is_symbolic || s.is_symbolic) { // if symbolic compare number of params
        equal = (equal && (this->param_names.size() == s.param_names.size()));
    } else { // if grounded compare params content
        equal = (equal && utils::compare_ordered_vector_strings(this->param_names, s.param_names));
    }
    return equal;
}

bool TypedSymbol::symbolic_compare(const TypedSymbol &s) const
{
    Symbol s1(this->name, this->param_names);
    Symbol s2(s.name, s.param_names);
    bool equal = (s1.symbolic_compare(s2));
    equal = (equal && utils::compare_ordered_vector_strings(this->param_types, s.param_types));
    return equal;
}

bool operator<(const Symbol& l, const Symbol& r)
{
    if (l.name < r.name) {
        return true;
    } else if (r.name < l.name) {
        return false;
    } else { //equal
        if (l.param_names.size() < r.param_names.size())
            return true;
        else if (r.param_names.size() < l.param_names.size())
            return false;
        else { // equal number of parameters
            std::vector< std::string >::const_iterator l_it, r_it;
            r_it = r.param_names.begin();
            for (l_it = l.param_names.begin(); l_it != l.param_names.end(); ++l_it) {
                if (*l_it < *r_it) {
                    return true;
                } else if (*r_it < *l_it) {
                    return false;
                }
                ++r_it;
            }
        }
    }
    return false;
}

bool operator<(const TypedSymbol& l, const TypedSymbol& r)
{
    if (l.name < r.name) {
        return true;
    } else if (r.name < l.name) {
        return false;
    } else { //equal
        // check param types
        if (l.param_types.size() < r.param_types.size())
            return true;
        else if (r.param_types.size() < l.param_types.size())
            return false;
        else { // equal number of parameter types
            std::vector< std::string >::const_iterator l_it, r_it;
            r_it = r.param_types.begin();
            for (l_it = l.param_types.begin(); l_it != l.param_types.end(); ++l_it) {
                if (*l_it < *r_it) {
                    return true;
                } else if (*r_it < *l_it) {
                    return false;
                }
                ++r_it;
            }
            // if it gets here, then everything has been equal
            
            // continue comparing parameter values
            if (l.param_names.size() < r.param_names.size())
                return true;
            else if (r.param_names.size() < l.param_names.size())
                return false;
            else { // equal number of parameters
                std::vector< std::string >::const_iterator l_it, r_it;
                r_it = r.param_names.begin();
                for (l_it = l.param_names.begin(); l_it != l.param_names.end(); ++l_it) {
                    if (*l_it < *r_it) {
                        return true;
                    } else if (*r_it < *l_it) {
                        return false;
                    }
                    ++r_it;
                }
            }
        }
    }
    return false;
}

std::ostream& operator<<(std::ostream &out, const Symbol &s)
{
    out << s.name << "(";
    for (size_t i = 0; i < s.param_names.size(); i++) {
        out << s.param_names[i];
        if ((i + 1) < s.param_names.size())
            out << " ";
    }
    out << ")";
    return out;
}

std::ostream& operator<<(std::ostream &out, const TypedSymbol &s)
{
    out << s.name << "(";
    s.write_typed_parameters(out);
    out << ")";
    return out;
}

void TypedSymbol::write_typed_parameters(std::ostream &out) const
{
    for (size_t i = 0; i < param_names.size(); i++) {
        out << param_names[i];
        out << " - ";
        out << param_types[i];
        if ((i + 1) < param_names.size())
            out << " ";
    }
}

void Symbol::write_PPDDL(std::ostream &out) const
{
    out << "(" << name;
    for (std::vector<std::string>::const_iterator param_it = param_names.begin(); param_it != param_names.end(); ++param_it) {
        out << " " << *param_it;
    }
    out << ")";
}

void TypedSymbol::write_PPDDL(std::ostream &out) const
{
    out << "(" << name << " ";
    write_typed_parameters(out);
    out << ")";
}


// utils
bool symbolic::is_param_symbolic(std::string param) 
{
    return (param[0] =='?');
}


// Symbol
Symbol::Symbol(const Symbol &symbol) :
    name(symbol.name),
    param_names(symbol.param_names),
    is_symbolic(symbol.is_symbolic)
{ }

Symbol::Symbol(const std::string &name_, const std::vector<std::string> &param_names_) :
    name(name_),
    param_names(param_names_)
{
    autodetect_symbolic();
}

Symbol::Symbol(const std::string &line):
    name(get_symbol_name_from_string(line)),
    param_names(get_symbol_param_names_from_string(line))
{
    autodetect_symbolic();
}

void Symbol::autodetect_symbolic()
{
    is_symbolic = false;
    for (vector_string::const_iterator param_it = param_names.begin(); param_it != param_names.end(); ++param_it) {
        if (symbolic::is_param_symbolic(*param_it)) { // found!
            is_symbolic = true;
            return;
        }
    }
}

bool Symbol::has_param(const std::string &required_param) const
{
    bool result = false;
    for (size_t i = 0; i < param_names.size(); i++) {
        if (param_names[i] == required_param) {
            result = true;
        }
    }
    return result;
}

int Symbol::get_param_name_idx(const std::string &param) const
{
    int result = -1;
    for (size_t i = 0; i < param_names.size(); i++) {
        if (param_names[i] == param) {
            result = i;
        }
    }
    return result;
}

bool Symbol::params_belong_to(const std::vector<std::string> &list_of_params) const
{
    bool result = true;
    for (size_t i = 0; i < list_of_params.size(); i++) {
        bool present = false;
        for (size_t j = 0; j < list_of_params.size(); j++) {
            if (param_names[i] == list_of_params[j]) {
                present = true;
            }
        }
        if (!present) {
            result = false;
        }
    }
    return result;
}
/*
bool Symbol::exists() const
{
    return (name.compare("") != 0);
}*/

bool Symbol::is_valid() const
{
    return (name != "");
}

void Symbol::make_symbolic()
{
    for (size_t i = 0; i < param_names.size(); ++i) {
        param_names[i] = DEFAULT_PARAM_NAMES[i];
    }
    is_symbolic = true;
}

bool Symbol::is_special() const
{
    bool result = ((*this == MAKE_SYMBOL_DEAD) || (*this == MAKE_SYMBOL_RETRY) || (*this == MAKE_SYMBOL_HELP) || (*this == MAKE_SYMBOL_UNKNOWN) || (*this == MAKE_SYMBOL_FINISH_ACTION) || (*this == MAKE_SYMBOL_TEACHER) || (*this == MAKE_SYMBOL_FINISHED));
    return result;
}

void Symbol::transform_nid_params_to_pddl_params()
{
    BOOST_FOREACH(std::string& param, param_names){
        if (param == "X") {
            param = "?X";
        }
        else if (param == "Y") {
            param = "?Y";
        }
        else if (param == "Z") {
            param = "?Z";
        }
        else if (param == "W") {
            param = "?W";
        }
        else if (param == "W") {
            param = "?W";
        }
        else if (param == "Q") {
            param = "?Q";
        }
    }
}

void Symbol::ground(std::vector<std::string> variables)
{
    std::random_shuffle(variables.begin(), variables.end());
    for (size_t i = 0; i < param_names.size(); ++i) {
        if (variables.size() > i) {
            param_names[i] = variables[i];
        } else {
            std::cout << "\033[1;31m" << "Symbols:: ground - Incorrect number of parameters." << "\033[0m" << std::endl;
            std::cout << "\033[1;31m" << "Symbol params are: " << param_names << "\033[0m" << std::endl;
            std::cout << "\033[1;31m" << "Arguments are: " << variables << "\033[0m" << std::endl;
        }
    }
}

void Symbol::PPDDL_grounding()
{
    if (param_names.size() > 0) // first parameter has two _
        name = name + "_";
    for (std::vector<std::string>::iterator param_it = param_names.begin(); param_it != param_names.end(); ++param_it) {
        name = name + "_" + *param_it;
    }
    param_names.clear();
}

/*
Symbol get_symbol_from_string(std::string line)
{
//     line = line.substr(line.find_first_not_of(" \t"));
    line = utils::trim(line);
    Symbol symbol(get_symbol_name_from_string(line), get_symbol_param_names_from_string(line));
    return symbol;
}*/

std::string Symbol::get_symbol_name_from_string(const std::string &line)
{
    std::stringstream ss(line);
    std::string item;

    if (std::getline(ss, item, '(')) {
        //std::cout << "Symbol name: " << item << std::endl;
    }

    return item;
}

std::vector< std::string > Symbol::get_symbol_param_names_from_string(const std::string &line)
{
    if (line.find('(') == std::string::npos) { // doesn't have params
        return std::vector< std::string >();
    }
    
    std::stringstream ss(line);
    std::string item;

    // get string between ()
    std::getline(ss, item, '(');
    std::getline(ss, item, ')');
    
    if ( (item.find('?') != std::string::npos) && (item.find(" - ") != std::string::npos)) {
        // if has '?', then it should be PPDDL rules
        return get_param_names_from_ppddl_parameters_string(item);
    }
    else {
        return get_param_names_from_nid_parameters_string(item);
    }
}

std::vector< std::string > Symbol::get_param_names_from_nid_parameters_string(const std::string &nid_parameters)
{
    std::stringstream ss;
    std::string item;
    std::vector< std::string> param_names;
    ss.str(nid_parameters);
    
    char separator = utils::choose_variable_separator(nid_parameters);
    while (std::getline(ss, item, separator)) {
        // remove spaces and empty strings
        item.erase(remove_if(item.begin(), item.end(), isspace), item.end());
        if (item.compare("") != 0) {
            param_names.push_back(item);
//             std::cout << "Param names: " << item << std::endl;
        }
    }

    return param_names;
}

std::vector< std::string > Symbol::get_param_names_from_ppddl_parameters_string(const std::string &ppddl_parameters)
{
    std::stringstream ss;
    std::string item;
    std::vector< std::string> param_names;
    ss.str(ppddl_parameters);
    
    return get_typed_params_from_ppddl_parameters_string(ppddl_parameters).front();
}

std::vector< std::vector<std::string> > Symbol::get_typed_params_from_ppddl_parameters_string(const std::string &ppddl_parameters)
{
    std::stringstream ss;
    std::string item;
    ss.str(ppddl_parameters);
    std::vector< std::vector<std::string> > typed_params; // typed_params[0] -> params, typed_params[1] -> types
    typed_params.push_back(std::vector<std::string>());
    typed_params.push_back(std::vector<std::string>());
    
    // get elements between ( )
    std::vector<std::string> elements;
    char separator = ' ';
    while (std::getline(ss, item, separator)) {
        // remove spaces and empty strings
        item.erase(remove_if(item.begin(), item.end(), isspace), item.end());
        if (item.compare("") != 0) {
            elements.push_back(item);
            //std::cout << "Param names: " << item << std::endl;
        }
    }
    
    for(size_t i = 0; i + 2 < elements.size(); i=i+3) {
        typed_params[0].push_back(elements[i]); // remove leading '?'
        // typed_params[0].push_back(elements[i].substr(1)); // remove leading '?'
        // elements[i+1] == '-'
        typed_params[1].push_back(elements[i+2]);
    }
    
    return typed_params;
}




/**
 * TYPED SYMBOL
 */

TypedSymbol::TypedSymbol(const TypedSymbol &symbol):
    Symbol(symbol),
    param_types(symbol.param_types)
{ 
    assert(param_names.size() == param_types.size());
}

TypedSymbol::TypedSymbol(const std::string &line):
    Symbol(line)
{
    load_typed_param_names_from_string(line);
    autodetect_symbolic();
    assert(param_names.size() == param_types.size());
}


TypedSymbol::TypedSymbol(const std::string &name_, const std::vector<std::string> &param_names_, const std::vector<std::string> &param_types_):
    Symbol(name_,param_names_),
    param_types(param_types_)
{
    assert(param_names.size() == param_types.size());
}


void TypedSymbol::load_typed_param_names_from_string(const std::string &line)
{
    if (line.find('(') == std::string::npos) { // doesn't have params
        return;
    }
    
    std::stringstream ss(line);
    std::string item;

    // get string between ()
    std::getline(ss, item, '(');
    std::getline(ss, item, ')');
    ss.str(item);
    
    if ( (item.find('?') == std::string::npos) && (item.find(" - ") == std::string::npos) ) {
        param_names = get_param_names_from_nid_parameters_string(item);
        for (size_t i = 0; i < param_names.size(); ++i) {
            param_types.push_back(OBJECT_TYPE_DEFAULT);
        }
        // HACK log once
        static bool hacky_bool = true;
        if (!param_names.empty()) {
            LOG_IF(hacky_bool,WARNING) << "NID compatibility, DEPRECATED. Reading: " << line;
        }
        hacky_bool = false;
    }
    else {
        std::vector< std::vector< std::string > > obtained_typed_params = get_typed_params_from_ppddl_parameters_string(item);
        param_names = obtained_typed_params[0];
        param_types = obtained_typed_params[1];
    }
}


bool TypedSymbol::is_special() const
{
    bool result = ((*this == MAKE_TYPED_SYMBOL_DEAD) || 
                   (*this == MAKE_TYPED_SYMBOL_RETRY) || 
                   (*this == MAKE_TYPED_SYMBOL_HELP) || 
                   (*this == MAKE_TYPED_SYMBOL_UNKNOWN) || 
                   (*this == MAKE_TYPED_SYMBOL_FINISH_ACTION) || 
                   (*this == MAKE_TYPED_SYMBOL_TEACHER));
    return result;
}



/**
 * PDDL objects and types manager
 */
void PDDLObjectManager::set_ppddl_objects_from_object_type_pairs(const std::vector<std::string>& object_names, const std::vector<std::string>& object_types)
{
    std::vector<std::string>::const_iterator object_names_it = object_names.begin();
    std::vector<std::string>::const_iterator object_types_it = object_types.begin();
    for ( ; object_names_it != object_names.end(); ++object_names_it) {
        ppddl_objects_types[*object_names_it] = *object_types_it;
        ppddl_objects_per_type[*object_types_it].push_back(*object_names_it);
        
        ++object_types_it;
    }
}

std::string PDDLObjectManager::get_object_type(const std::string& object) const
{
    PDDLObjectTypeMap::const_iterator iter = ppddl_objects_types.find(object);
    if (iter != ppddl_objects_types.end()) {
        return iter->second;
    }
    else {
        return OBJECT_TYPE_DEFAULT;
    }
}

std::vector<std::string> PDDLObjectManager::get_type_objects(const std::string& type) const
{
    PDDLTypeObjectMap::const_iterator iter = ppddl_objects_per_type.find(type);
    if (iter != ppddl_objects_per_type.end()) {
        return iter->second;
    }
    else {
        return std::vector<std::string>();
    }
}

TypedSymbol PDDLObjectManager::get_typed_symbol(const Symbol& symbol) const
{
    std::vector<std::string> action_param_types;
    BOOST_FOREACH(const std::string& param_name, symbol.param_names) {
        action_param_types.push_back(get_object_type(param_name));
    }
    return TypedSymbol(symbol.name, symbol.param_names, action_param_types);
}
