/*
 * Class representing a logic symbol with parameters.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef IRI_RULE_LEARNER_SYMBOLS_H
#define IRI_RULE_LEARNER_SYMBOLS_H 1

#include <algorithm>
#include <vector>
#include <set>
#include <fstream>
#include <sstream>

#include <boost/optional.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include "symbolic/utils.h"

// Typed objects
#define OBJECT_TYPE_DEFAULT "default"


// actions
#define MAKE_SYMBOL_DEAD          Symbol("dead()")
#define MAKE_SYMBOL_RETRY         Symbol("retry()")
#define MAKE_SYMBOL_HELP          Symbol("help()")
#define MAKE_SYMBOL_UNKNOWN       Symbol("unknownManipulation()")
#define MAKE_SYMBOL_TEACHER       Symbol("teacher()")
#define MAKE_SYMBOL_FINISH_ACTION Symbol("finish_action()")

#define MAKE_TYPED_SYMBOL_DEAD          TypedSymbol("dead", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_RETRY         TypedSymbol("retry", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_HELP          TypedSymbol("help", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_UNKNOWN       TypedSymbol("unknownManipulation", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_TEACHER       TypedSymbol("teacher", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_FINISH_ACTION TypedSymbol("finish_action", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_NOOP_ACTION TypedSymbol("noop", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_SYMBOL_DEFAULT       TypedSymbol("default", std::vector<std::string>(), std::vector<std::string>())

// predicates
#define MAKE_SYMBOL_FINISHED         Symbol("finished()")
#define MAKE_SYMBOL_STARTED          Symbol("started()")
#define MAKE_SYMBOL_EXCUSE_GENERATED Symbol("excuseGenerated()")

// action_ptr
#define MAKE_SYMBOL_DEAD_PTR    boost::make_shared<Symbol>(MAKE_SYMBOL_DEAD)
#define MAKE_SYMBOL_RETRY_PTR   boost::make_shared<Symbol>(MAKE_SYMBOL_RETRY)
#define MAKE_SYMBOL_HELP_PTR    boost::make_shared<Symbol>(MAKE_SYMBOL_HELP)
#define MAKE_SYMBOL_UNKNOWN_PTR boost::make_shared<Symbol>(MAKE_SYMBOL_UNKNOWN)
#define MAKE_SYMBOL_TEACHER_PTR boost::make_shared<Symbol>(MAKE_SYMBOL_TEACHER)

static const std::string DEFAULT_PARAM_NAMES[] = {"?X", "?Y", "?Z", "?V", "?W", "?Q"}; //!< Default param names that are considered symbolic.
static const uint NUMBER_DEFAULT_PARAM_NAMES = 6;
static const std::string DEFAULT_ACTION_NAME = "default";
static std::vector<std::string> DEFAULT_PARAM_NAMES_VECTOR(DEFAULT_PARAM_NAMES, DEFAULT_PARAM_NAMES + NUMBER_DEFAULT_PARAM_NAMES);



namespace symbolic {
    bool is_param_symbolic(std::string param);
}


/*!
\class Symbol
\brief Class representing a symbol. Symbols consist of a name and a vector of parameters.
*/
class Symbol
{
public:
    std::string name; //!< Name
    std::vector<std::string> param_names; //!< description Vector of parameters
    bool is_symbolic; //!< true if it is symbolic, false if grounded

    /*!
    \brief Empty constructor
    */
    Symbol() {
        autodetect_symbolic();
    }

    /*!
    \brief Copy constructor
    */
    Symbol(const Symbol &symbol);

    /*!
    \brief constructor

    Parses a string that is assumed to have a symbol.
    Examples of the syntax supported:
     "clear(X Y)", "clear(X,Y)", "clear()".

    \param line string having a valid representation of a symbol
    */
    Symbol(const std::string &line);
    /*!
    \brief Constructor

    Constructs the symbol.

    \param name_ name
    \param param_names_ vector of param names
    */
    Symbol(const std::string &name_, const std::vector<std::string> &param_names_);

    /*!
    \brief Checks if symbol is symbolic, and updates its is_symbolic member

    If one of the default param names (X, Y, Z, ...) is found,
    is_symbolic is set to true.
    */
    void autodetect_symbolic();
    
    /*!
    \brief Check if symbol is a special symbol
    
    Checks if symbol is a special symbol: dead(), retry(), help(), unknown(), finish(), teacher().

    \return true if it is special
    */
    bool is_special() const;
    
    /*!
    \brief Check if a parameter with the given name exists

    \return true if there is a parameter with the given name
    */
    bool has_param(const std::string &required_param) const;

    /*!
    \brief Check if parameter with all the given name exist

    \return true if all of them exist
    */
    bool params_belong_to(const std::vector<std::string> &list_of_params) const;

    /*!
    \brief Check if symbol is initialized

    Check if symbol is initialized, that is that its name is not empty.

    \return true if initialized, false if empty
    */
    bool is_valid() const;

    /*!
    \brief Gets the index of a param

    \param param name of the param to look for its index
    \return true if it is special
    */
    int get_param_name_idx(const std::string &param) const;

    /*!
    \brief Makes the symbol symbolic

    Changes the params to be the default symbolic params: X, Y, Z, V, W, Q.
    */
    void make_symbolic();

    /*!
    \brief Grounds the symbol with the given parameters
    */
    void ground(std::vector<std::string> variables);

    /*!
    \brief Makes parameter-less symbol appending parameters to the name
    */
    void PPDDL_grounding();

    /*!
    \brief Writes PPDDL grounded predicates without parameters
    */
    void write_PPDDL(std::ostream &out) const;

    /*!
    \brief Substitutes X for ?X, Y for ?Y, etc...
    */
    void transform_nid_params_to_pddl_params();

    /*!
    \brief Compares if not grounded versions of the symbol are equal

    Compares the name and the number of parameters, but not the parameters content

    \return true if not grounded symbols are equal
    */
    bool symbolic_compare(const Symbol &s) const;

    /*! \brief Operator == */
    friend bool operator== (const Symbol &s1, const Symbol &s2);
    /*! \brief Operator != */
    friend bool operator!= (const Symbol &s1, const Symbol &s2);
    /*! \brief Operator <= */
    friend bool operator<(const Symbol& l, const Symbol& r);
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const Symbol &s);

protected:
    static std::string get_symbol_name_from_string(const std::string &line);
    static std::vector< std::string > get_symbol_param_names_from_string(const std::string &line);
    static std::vector< std::string > get_param_names_from_nid_parameters_string(const std::string &nid_parameters);
    static std::vector< std::string > get_param_names_from_ppddl_parameters_string(const std::string &ppddl_parameters);
    static std::vector< std::vector<std::string> > get_typed_params_from_ppddl_parameters_string(const std::string &ppddl_parameters);
};

/*! Optional symbol, it is false if it doesn't exist */
typedef boost::optional<Symbol> SymbolOptional;
/*! Shared pointer to symbol */
typedef boost::shared_ptr<const Symbol> SymbolPtr;
/*! Shared pointer to const symbol */
typedef boost::shared_ptr<const Symbol> SymbolConstPtr;



class TypedSymbol : public Symbol
{
public:
    std::vector< std::string > param_types; //!< Types of the parameters

    /*!
    \brief Empty constructor
    */
    TypedSymbol() : Symbol() { }
    
    /*!
    \brief Copy constructor
    */
    TypedSymbol(const TypedSymbol &typed_symbol);

    /*!
    \brief constructor

    Parses a string that is assumed to have a symbol.
    Examples of the syntax supported:
     "clear(?X - location ?Y - thing)", "clear()".

    \param line string having a valid representation of a symbol
    */
    TypedSymbol(const std::string &line);

    /*!
    \brief Constructor

    Constructs the symbol.

    \param name_ name
    \param param_names_ vector of param names
    */
    TypedSymbol(const std::string &name_, const std::vector<std::string> &param_names_, const std::vector<std::string> &param_types_);

    /*!
    \brief Compares if not grounded versions of the symbol are equal

    Compares the name and the number of parameters, but not the parameters content

    \return true if not grounded symbols are equal
    */
    bool symbolic_compare(const TypedSymbol &s) const;

    /*!
    \brief Check if TypedSymbol is a special symbol

    Checks if symbol is a special symbol: dead(), retry(), help(), unknown(), finish(), teacher().

    \return true if it is special
    */
    bool is_special() const;
    
    /*!
    \brief Writes PPDDL typed symbol
    */
    void write_PPDDL(std::ostream &out) const;

    /*! \brief Write typed params */
    void write_typed_parameters(std::ostream &out) const;
    
    /*! \brief Operator == */
    friend bool operator== (const TypedSymbol &s1, const TypedSymbol &s2);
    /*! \brief Operator != */
    friend bool operator!= (const TypedSymbol &s1, const TypedSymbol &s2);
    /*! \brief Operator <= */
    friend bool operator<(const TypedSymbol& l, const TypedSymbol& r);
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const TypedSymbol &s);

private:
    void load_typed_param_names_from_string(const std::string &line);
};


typedef std::string PDDLType;
typedef std::vector< std::string > PDDLTypeVector;
typedef std::map< PDDLType, std::vector< std::string > > PDDLTypeObjectMap;
typedef std::map< std::string, PDDLType > PDDLObjectTypeMap;


/*!
\class PDDLObjectManager
\brief Class to manage objects and types.
*/
class PDDLObjectManager
{
public:
    PDDLObjectManager() {};
    
    PDDLObjectManager(const PDDLObjectManager& other):
        ppddl_objects_per_type(other.ppddl_objects_per_type),
        ppddl_objects_types(other.ppddl_objects_types)
    {};
    
    /**
     * \brief Gets typed symbol by searching is corresponding param types
     * 
     * TODO C++11 const method -> use map::at() which throws exceptions
     */
    TypedSymbol get_typed_symbol(const Symbol& symbol) const;
    
    /**
     * \brief Gets the type of the object
     * 
     * TODO C++11 const method -> use map::at() which throws exceptions
     */
    std::string get_object_type(const std::string& object) const;
    
    /**
     * \brief Gets the objects with type type
     * 
     * TODO C++11 const method -> use map::at() which throws exceptions
     */
    std::vector<std::string> get_type_objects(const std::string& type) const;
    
    /**
     * \brief Gets vectors of ppddl objects associated to each type
     */
    const PDDLTypeObjectMap& get_ppddl_objects_per_type() const { return ppddl_objects_per_type; };
    
    /**
     * \brief Gets the type associated to each object
     */
    const PDDLObjectTypeMap& get_ppddl_objects_types() const { return ppddl_objects_types; };
    
    /**
     * \brief Get list of objects
     */
    std::vector<std::string> get_ppddl_objects() const { 
        PDDLTypeVector res; 
        BOOST_FOREACH(PDDLObjectTypeMap::value_type object_type, ppddl_objects_types) {
            res.push_back(object_type.first);
        } 
        return res; 
    };
    
    /**
     * \brief Set objects from a vector of object-type pairs
     */
    void set_ppddl_objects_from_object_type_pairs(const std::vector<std::string>& object_names, const std::vector<std::string>& object_types);
    
private:
    PDDLTypeObjectMap ppddl_objects_per_type;
    PDDLObjectTypeMap ppddl_objects_types;
};


#endif
